# -*- coding: utf-8 -*-


import re

from widlparser import parser
from . import config
from .html import *
from .messages import *
from .unsortedJunk import classifyDfns


class IDLUI(object):
    def warn(self, msg):
        die("{0}", msg.rstrip())


class IDLSilent(object):
    def warn(self, msg):
        pass


class DebugMarker(object):
    # Debugging tool for IDL markup

    def markup_construct(self, text, construct):
        return ('<construct-' + construct.idl_type + '>', '</construct-' + construct.idl_type + '>')

    def markup_type(self, text, construct):
        return ('<TYPE for="' + construct.idl_type + '" idlType="' + text + '">', '</TYPE>')

    def markup_primitive_type(self, text, construct):
        return ('<PRIMITIVE for="' + construct.idl_type + '" idlType="' + text + '">', '</PRIMITIVE>')

    def markup_buffer_type(self, text, construct):
        return ('<BUFFER for="' + construct.idl_type + '" idlType="' + text + '">', '</BUFFER>')

    def markup_string_type(self, text, construct):
        return ('<STRING for="' + construct.idl_type + '" idlType="' + text + '">', '</STRING>')

    def markup_object_type(self, text, construct):
        return ('<OBJECT for="' + construct.idl_type + '" idlType="' + text + '">', '</OBJECT>')

    def markup_type_name(self, text, construct):
        return ('<TYPE-NAME idlType="' + construct.idl_type + '">', '</TYPE-NAME>')

    def markup_name(self, text, construct):
        return ('<NAME idlType="' + construct.idl_type + '">', '</NAME>')

    def markup_keyword(self, text, construct):
        return ('<KEYWORD idlType="' + construct.idl_type + '">', '</KEYWORD>')

    def markup_enum_value(self, text, construct):
        return ('<ENUM-VALUE for="' + construct.name + '">', '</ENUM-VALUE>')


class IDLMarker(object):
    def markup_construct(self, text, construct):
        # Fires for every 'construct' in the WebIDL.
        # Some things are "productions", not "constructs".
        return (None, None)

    def markup_type(self, text, construct):
        # Fires for entire type definitions.
        # It'll contain keywords or names, or sometimes more types.
        # For example, a "type" wrapper surrounds an entire union type,
        # as well as its component types.
        return (None, None)

    def markup_primitive_type(self, text, construct):
        return ("<a data-link-type=interface>", "</a>")

    def markup_string_type(self, text, construct):
        return ("<a data-link-type=interface>", "</a>")

    def markup_buffer_type(self, text, construct):
        return ("<a data-link-type=interface>", "</a>")

    def markup_object_type(self, text, construct):
        return ("<a data-link-type=interface>", "</a>")

    def markup_type_name(self, text, construct):
        # Fires for non-defining type names, such as arg types.

        # The names in [Exposed=Foo] are [Global] tokens, not interface names.
        # Since I don't track globals as a link target yet, don't link them at all.
        if construct.idl_type == "extended-attribute" and construct.name == "Exposed":
            return (None, None)

        # The name in [PutForwards=foo] is an attribute of the same interface.
        if construct.idl_type == "extended-attribute" and construct.name == "PutForwards":
            # In [PutForwards=value] attribute DOMString foo
            # the "value" is a DOMString attr
            attr = construct.parent
            if hasattr(attr.member, "rest"):
                type = attr.member.rest.type
            elif hasattr(attr.member, "attribute"):
                type = attr.member.attribute.type
            typeName = str(type).strip()
            if typeName.endswith("?"):
                typeName = typeName[:-1]
            return ('<a data-link-type=attribute data-link-for="{0}">'.format(typeName), '</a>')

        # LegacyWindowAlias defines additional names for the construct,
        # so all the names should be forced <dfn>s, just like the interface name itself.
        if construct.idl_type == "extended-attribute" and construct.name == "LegacyWindowAlias":
            return ('<idl data-idl-type=interface data-lt="{0}">'.format(text), '</idl>')

        if construct.idl_type == "constructor":
            # This shows up for the method name in a [NamedConstructor] extended attribute.
            # The "NamedConstructor" Name already got markup up, so ignore this one.
            return (None, None)

        return ('<a class=n data-link-type="idl-name">', '</a>')

    def markup_keyword(self, text, construct):
        # Fires on the various "keywords" of WebIDL -
        # words that are part of the WebIDL syntax,
        # rather than names exposed to JS.
        # Examples: "interface", "stringifier", the IDL-defined type names like "DOMString" and "long".
        if text == "stringifier":
            if construct.name is None:
                # If no name was defined, you're required to define stringification behavior.
                return ("<a dfn for='{0}' data-lt='stringification behavior'>".format(construct.parent.full_name), "</a>")
            else:
                # Otherwise, you *can* point to/dfn stringification behavior if you want.
                return ("<idl data-export data-idl-type=dfn data-idl-for='{0}' data-lt='stringification behavior' id='{0}-stringification-behavior'>".format(construct.parent.full_name), "</idl>")
        return (None, None)

    def markup_name(self, text, construct):
        # Fires for defining names: method names, arg names, interface names, etc.
        idlType = construct.idl_type
        if idlType not in config.idlTypes:
            return (None, None)

        if idlType == "constructor":
            # the [Constructor] extended attr, now deprecated
            die(f"The [Constructor] extended attribute (on {construct.parent.name}) is deprecated, please switch to a constructor() method.")
            return (None, None)

        if idlType == "argument" and construct.parent.idl_type == "constructor":
            # Don't mark up the arguments to [Constructor] either
            return (None, None)

        idlTitle = construct.normal_name
        if idlType == "method" and idlTitle.startswith("constructor("):
            idlType = "constructor"

        extraParameters = ''
        refType = "idl"
        if idlType in config.functionishTypes:
            idlTitle = '|'.join(self.methodLinkingTexts(construct))
        elif idlType == "extended-attribute":
            refType = "link"
        elif idlType == "attribute":
            if hasattr(construct.member, "rest"):
                rest = construct.member.rest
            elif hasattr(construct.member, "attribute"):
                rest = construct.member.attribute
            else:
                die("Can't figure out how to construct attribute-info from:\n  {0}", construct)
            if rest.readonly is not None:
                readonly = 'data-readonly'
            else:
                readonly = ''
            extraParameters = '{0} data-type="{1}"'.format(readonly, str(rest.type).strip())
        elif idlType == "dict-member":
            extraParameters = 'data-type="{0}"'.format(construct.type)
            if construct.default is not None:
                value = str(construct.default).split('=', 1)[1].strip()
                if value.startswith('['):
                    value = '[]'
                elif value.startswith('}'):
                    value = '{}'
                extraParameters += ' data-default="{0}"'.format(escapeAttr(value))
        elif idlType in ["interface", "namespace", "dictionary"]:
            if construct.partial:
                refType = "link"

        if refType == "link":
            elementName = "a"
        else:
            elementName = "idl"

        if idlType in config.typesUsingFor:
            if idlType == "argument" and construct.parent.idl_type == "method":
                interfaceName = construct.parent.parent.name
                methodNames = ["{0}/{1}".format(interfaceName, m) for m in self.methodLinkingTexts(construct.parent)]
                idlFor = "data-idl-for='{0}'".format(", ".join(methodNames))
            else:
                idlFor = "data-idl-for='{0}'".format(construct.parent.full_name)
        else:
            idlFor = ""
        return ('<{name} data-lt="{0}" data-{refType}-type="{1}" {2} {3}>'.format(idlTitle, idlType, idlFor, extraParameters, name=elementName, refType=refType), '</{0}>'.format(elementName))

    def markup_enum_value(self, text, construct):
        return ("<idl data-idl-type=enum-value data-idl-for='{0}' data-lt='{1}'>".format(escapeAttr(construct.name), escapeAttr(text)), "</idl>")

    def encode(self, text):
        return escapeHTML(text)

    def methodLinkingTexts(self, method):
        '''
        Given a method-ish widlparser Construct,
        finds all possible linking texts.
        The full linking text is "foo(bar, baz)";
        beyond that, any optional or variadic arguments can be omitted.
        So, if both were optional,
        "foo(bar)" and "foo()" would both also be valid linking texts.
        '''
        for i,arg in enumerate(method.arguments or []):
            if arg.optional or arg.variadic:
                optStart = i
                break
        else:
            optStart = None

        texts = []

        if optStart is not None:
            prefixes = [method.name]
            if method.name == "constructor":
                prefixes.append(method.parent.name)
            for i in range(optStart, len(method.arguments)):
                argText = ', '.join(arg.name for arg in method.arguments[:i])
                for prefix in prefixes:
                    texts.append(prefix + "(" + argText + ")")

        texts.append(method.normal_name)
        if method.name == "constructor":
            texts.append(method.parent.name + method.normal_name[11:])
        return reversed(texts)


def markupIDL(doc):
    highlightingOccurred = False
    idlEls = findAll("pre.idl:not([data-no-idl]), xmp.idl:not([data-no-idl])", doc)
    for el in findAll("script[type=idl]", doc):
        # To help with syntax-highlighting, <script type=idl> is also allowed here.
        idlEls.append(el)
        el.tag = "pre"
        removeAttr(el, "type")
        addClass(el, "idl")
    # One pass with a silent parser to collect the symbol table.
    symbolTable = None
    for el in idlEls:
        p = parser.Parser(textContent(el), ui=IDLSilent(), symbol_table=symbolTable)
        symbolTable = p.symbol_table
    # Then a real pass to actually mark up the IDL,
    # and collect it for the index.
    for el in idlEls:
        if isNormative(el, doc):
            text = textContent(el)
            # Parse once with a fresh parser, so I can spit out just this <pre>'s markup.
            widl = parser.Parser(text, ui=IDLUI(), symbol_table=symbolTable)
            marker = DebugMarker() if doc.debug else IDLMarker()
            replaceContents(el, parseHTML(str(widl.markup(marker))))
            # Parse a second time with the global one, which collects all data in the doc.
            doc.widl.parse(text)
        addClass(el, "highlight")
        highlightingOccurred = True
    if doc.md.slimBuildArtifact:
        # Remove the highlight-only spans
        for el in idlEls:
            for span in findAll("span", el):
                contents = childNodes(span, clear=True)
                replaceNode(span, *contents)
        return
    if highlightingOccurred:
        doc.extraStyles['style-syntax-highlighting'] += "pre.idl.highlight { color: #708090; }"


def processIDL(doc):
    localDfns = set()
    for pre in findAll("pre.idl, xmp.idl", doc):
        if pre.get("data-no-idl") is not None:
            continue
        if not isNormative(pre, doc):
            continue
        forcedInterfaces = []
        for x in (treeAttr(pre, "data-dfn-force") or "").split():
            x = x.strip()
            if x.endswith("<interface>"):
                x = x[:-11]
            forcedInterfaces.append(x)
        for el in findAll("idl", pre):
            idlType = el.get('data-idl-type')
            url = None
            forceDfn = False
            ref = None
            for idlText in el.get('data-lt').split('|'):
                if idlType == "interface" and idlText in forcedInterfaces:
                    forceDfn = True
                for linkFor in config.splitForValues(el.get('data-idl-for', '')) or [None]:
                    ref = doc.refs.getRef(idlType, idlText,
                                          linkFor=linkFor,
                                          status="local",
                                          el=el,
                                          error=False)
                    if ref:
                        url = ref.url
                        break
                if ref:
                    break
            if url is None or forceDfn:
                el.tag = "dfn"
                el.set('data-dfn-type', idlType)
                del el.attrib['data-idl-type']
                if el.get('data-idl-for'):
                    el.set('data-dfn-for', el.get('data-idl-for'))
                    del el.attrib['data-idl-for']
            else:
                # Copy over the auto-generated linking text to the manual dfn.
                dfn = find(url, doc)
                lts = combineIdlLinkingTexts(el.get("data-lt"), dfn.get("data-lt"))
                dfn.set("data-lt", lts)
                localDfns.add(dfn)

                # Reset the <idl> element to be a link to the manual dfn.
                el.tag = "a"
                el.set('data-link-type', idlType)
                el.set('data-lt', idlText)
                del el.attrib['data-idl-type']
                if el.get('data-idl-for'):
                    el.set('data-link-for', el.get('data-idl-for'))
                    del el.attrib['data-idl-for']
                if el.get('id'):
                    # ID was defensively added by the Marker.
                    del el.attrib['id']

    dfns = findAll("pre.idl:not([data-no-idl]) dfn, xmp.idl:not([data-no-idl]) dfn", doc) + list(localDfns)
    classifyDfns(doc, dfns)
    fixupIDs(doc, dfns)
    doc.refs.addLocalDfns(dfn for dfn in dfns if dfn.get('id') is not None)

def combineIdlLinkingTexts(t1, t2):
    t1s = [normalizeIdlWhitespace(x) for x in (t1 or "").split("|")]
    t2s = [normalizeIdlWhitespace(x) for x in (t2 or "").split("|")]
    for lt in t2s:
        if lt not in t1s:
            t1s.append(lt)
    return "|".join(t1s)

def normalizeIdlWhitespace(text):
    # Remove all whitespace...
    text = re.sub(r"\s+", "", text)
    # Then add whitespace after commas
    text = re.sub(r",", ", ", text)
    return text




def getParser():
    return parser.Parser(ui=IDLSilent())



import re
from ..html import E, outerHTML
from . import steps

class SectionShorthand:
    def __init__(self):
        self.stage = "start"
        self.escapedText = None
        self.linkText = []
        self.bsAutolink = ""

    def respond(self, match, dom=None):
        if self.stage == "start":
            return self.respondStart(match)
        elif self.stage == "link text":
            return self.respondLinkText(match, dom)
        elif self.stage == "end":
            return self.respondEnd()

    def respondStart(self, match):
        self.bsAutolink = match.group(0)
        escape, self.spec, self.section, self.justPage, hasLinkText = match.groups()
        if escape:
            self.escapedText = match.group(0)[1:]
        if hasLinkText:
            self.stage = "link text"
            return steps.NextBody(endRe)
        else:
            self.stage = "end"
            return steps.NextLiteral(endRe)

    def respondLinkText(self, match, dom):
        self.linkText = dom
        self.bsAutolink += outerHTML(dom)
        return self.respondEnd()

    def respondEnd(self):
        if self.escapedText:
            return steps.Success(skips=["["], nodes=[self.escapedText[1:], *self.linkText, "]]"])

        self.bsAutolink += "]]"

        if not self.linkText:
            self.linkText = "" # will get filled in by a later step

        if self.spec is None:
            # local section link
            attrs = {
                "section":"",
                "href":self.section,
                "bs-autolink-syntax":self.bsAutolink,
                }
            return steps.Success(E.a(attrs, self.linkText))
        elif self.justPage is not None:
            # foreign link, to an actual page from a multipage spec
            attrs = {
                "spec-section":self.justPage + "#",
                "spec":self.spec,
                "bs-autolink-syntax":self.bsAutolink,
                }
            return steps.Success(E.span(attrs, self.linkText))
        else:
            # foreign link
            attrs = {
                "spec-section":self.section,
                "spec":self.spec,
                "bs-autolink-syntax":self.bsAutolink,
                }
            return steps.Success(E.span(attrs, self.linkText))

SectionShorthand.startRe = re.compile(r"""
    (\\)?
    \[\[
    ([\w.+-]+)?
    (?:
        ((?:\/[\w.+-]*)?(?:\#[\w.+-]+)) |
        (\/[\w.+-]+)
    )
    (\|)?
    """, re.X)

endRe = re.compile(r"]]")



import re
from ..html import E, outerHTML
from . import steps

class BiblioShorthand:
    def __init__(self):
        self.stage = "start"
        self.escapedText = None
        self.linkText = []
        self.bsAutolink = ""

    def respond(self, match, dom=None):
        if self.stage == "start":
            return self.respondStart(match)
        elif self.stage == "link text":
            return self.respondLinkText(match, dom)
        elif self.stage == "end":
            return self.respondEnd()

    def respondStart(self, match):
        self.bsAutolink = match.group(0)
        escape, bang, self.term, self.status, hasLinkText = match.groups()
        if escape:
            self.escapedText = match.group(0)[1:]
        if bang == "!":
            self.type = "normative"
        else:
            self.type = "informative"
        if hasLinkText:
            self.stage = "link text"
            return steps.NextBody(biblioEndRe)
        else:
            self.stage = "end"
            return steps.NextLiteral(biblioEndRe)

    def respondLinkText(self, match, dom):
        self.linkText = dom
        self.bsAutolink += outerHTML(dom)
        return self.respondEnd()

    def respondEnd(self):
        if self.escapedText:
            return steps.Success(skips=["["], nodes=[self.escapedText[1:], *self.linkText, "]]"])

        self.bsAutolink += "]]"

        if not self.linkText:
            self.linkText = "[{0}]".format(self.term)

        attrs = {
            "data-lt":self.term,
            "data-link-type":"biblio",
            "data-biblio-type":self.type,
            "bs-autolink-syntax":self.bsAutolink}
        if self.status is not None:
            attrs['data-biblio-status'] = self.status.strip()

        return steps.Success(E.a(attrs, self.linkText))

BiblioShorthand.startRe = re.compile(r"""
    (\\)?
    \[\[
    (!)?
    ([\w.+-]+)
    (?:\s+(current|snapshot))?
    (\|)?""", re.X)

biblioEndRe = re.compile(r"]]")


# -*- coding: utf-8 -*-


import os

from . import updateBackRefs
from . import updateCrossRefs
from . import updateBiblio
from . import updateCanIUse
from . import updateMdn
from . import updateLinkDefaults
from . import updateTestSuites
from . import updateLanguages
from . import updateWpt
from . import manifest
from .. import config
from ..messages import *


def update(anchors=False, backrefs=False, biblio=False, caniuse=False, linkDefaults=False, mdn=False, testSuites=False, languages=False, wpt=False, path=None, dryRun=False, force=False):
    if path is None:
        path = config.scriptPath("spec-data")

    # Update via manifest by default, falling back to a full update only if failed or forced.
    if not force:
        success = manifest.updateByManifest(path=path, dryRun=dryRun)
        if not success:
            say("Falling back to a manual update...")
            force = True
    if force:
        # If all are False, update everything
        if  anchors == backrefs == biblio == caniuse == linkDefaults == mdn == testSuites == languages == wpt == False:
            anchors  = backrefs  = biblio  = caniuse  = linkDefaults  = mdn  = testSuites  = languages  = wpt  = True

        touchedPaths = {
            "anchors": updateCrossRefs.update(path=path, dryRun=dryRun) if anchors else None,
            "backrefs": updateBackRefs.update(path=path, dryRun=dryRun) if backrefs else None,
            "biblio": updateBiblio.update(path=path, dryRun=dryRun) if biblio else None,
            "caniuse": updateCanIUse.update(path=path, dryRun=dryRun) if caniuse else None,
            "mdnspeclinks": updateMdn.update(path=path, dryRun=dryRun) if mdn else None,
            "linkDefaults": updateLinkDefaults.update(path=path, dryRun=dryRun) if linkDefaults else None,
            "testSuites": updateTestSuites.update(path=path, dryRun=dryRun) if testSuites else None,
            "languages": updateLanguages.update(path=path, dryRun=dryRun) if languages else None,
            "wpt": updateWpt.update(path=path, dryRun=dryRun)if wpt else None
        }

        cleanupFiles(path, touchedPaths=touchedPaths, dryRun=dryRun)
        manifest.createManifest(path=path, dryRun=dryRun)



def fixupDataFiles():
    '''
    Checks the readonly/ version is more recent than your current mutable data files.
    This happens if I changed the datafile format and shipped updated files as a result;
    using the legacy files with the new code is quite bad!
    '''
    try:
        localVersion = int(open(localPath("version.txt"), 'r').read())
    except IOError:
        localVersion = None
    try:
        remoteVersion = int(open(remotePath("version.txt"), 'r').read())
    except IOError as err:
        warn("Couldn't check the datafile version. Bikeshed may be unstable.\n{0}", err)
        return

    if localVersion == remoteVersion:
        # Cool
        return

    # If versions don't match, either the remote versions have been updated
    # (and we should switch you to them, because formats may have changed),
    # or you're using a historical version of Bikeshed (ditto).
    try:
        for filename in os.listdir(remotePath()):
            copyanything(remotePath(filename), localPath(filename))
    except Exception as err:
        warn("Couldn't update datafiles from cache. Bikeshed may be unstable.\n{0}", err)
        return


def updateReadonlyDataFiles():
    '''
    Like fixupDataFiles(), but in the opposite direction --
    copies all my current mutable data files into the readonly directory.
    This is a debugging tool to help me quickly update the built-in data files,
    and will not be called as part of normal operation.
    '''
    try:
        for filename in os.listdir(localPath()):
            if filename.startswith("readonly"):
                continue
            copyanything(localPath(filename), remotePath(filename))
    except Exception as err:
        warn("Error copying over the datafiles:\n{0}", err)
        return


def cleanupFiles(root, touchedPaths, dryRun=False):
    if dryRun:
        return
    paths = set()
    deletableFiles = []
    deletableFolders = []
    if touchedPaths["anchors"] is not None:
        deletableFiles.extend(["specs.json", "methods.json", "fors.json"])
        deletableFolders.extend(["headings", "anchors"])
        paths.update(touchedPaths["anchors"])
    if touchedPaths["biblio"] is not None:
        deletableFiles.extend(["biblio-keys.json", "biblio-numeric-suffixes.json"])
        deletableFolders.extend(["biblio"])
        paths.update(touchedPaths["biblio"])
    if touchedPaths["caniuse"] is not None:
        deletableFiles.extend(["caniuse.json"])
        deletableFolders.extend(["caniuse"])
        paths.update(touchedPaths["caniuse"])
    if touchedPaths["mdnspeclinks"] is not None:
        deletableFolders.extend(["mdnspeclinks"])
        paths.update(touchedPaths["mdnspeclinks"])

    say("Cleaning up old data files...")
    oldPaths = []
    for absPath, relPath in getDatafilePaths(root):
        if "/" not in relPath and relPath not in deletableFiles:
            continue
        if "/" in relPath and relPath.partition("/")[0] not in deletableFolders:
            continue
        if absPath not in paths:
            os.remove(absPath)
            oldPaths.append(relPath)
    if oldPaths:
        say("Success! Deleted {0} old files.".format(len(oldPaths)))
    else:
        say("Success! Nothing to delete.")


def copyanything(src, dst):
    import shutil
    import errno
    try:
        shutil.rmtree(dst, ignore_errors=True)
        shutil.copytree(src, dst)
    except OSError as exc:
        if exc.errno in [errno.ENOTDIR, errno.EINVAL]:
            shutil.copy(src, dst)
        else:
            raise

def localPath(*segs):
    return config.scriptPath("spec-data", *segs)

def remotePath(*segs):
    return config.scriptPath("spec-data", "readonly", *segs)

def getDatafilePaths(basePath):
    for root, dirs, files in os.walk(basePath):
        for filename in files:
            filePath = os.path.join(root, filename)
            yield filePath, os.path.relpath(filePath, basePath)



# -*- coding: utf-8 -*-


import io
import itertools
import json
import logging
import os
import re
import urllib.parse, urllib.error
from collections import defaultdict, namedtuple
from functools import partial as curry

from . import biblio
from . import config
from . import datablocks
from . import dfnpanels
from . import func
from . import markdown
from .html import *
from .messages import *

class MarkdownCodeSpans(func.Functor):
    # Wraps a string, such that the contained text is "safe"
    # and contains no markdown code spans.
    # Thus, functions mapping over the text can freely make substitutions,
    # knowing they won't accidentally replace stuff in a code span.
    def __init__(self, text):
        self.__codeSpanReplacements__ = []
        newText = ""
        mode = "text"
        indexSoFar = 0
        backtickCount = 0
        for m in re.finditer(r"(\\`)|([\w-]*)(`+)", text):
            if mode == "text":
                if m.group(1):
                    newText += text[indexSoFar:m.start()] + m.group(1)[1]
                    indexSoFar = m.end()
                elif m.group(3):
                    mode = "code"
                    newText += text[indexSoFar:m.start()]
                    indexSoFar = m.end()
                    backtickCount = len(m.group(3))
                    tag = m.group(2)
                    literalStart = m.group(0)
            elif mode == "code":
                if m.group(1):
                    pass
                elif m.group(3):
                    if len(m.group(3)) != backtickCount:
                        pass
                    else:
                        mode = "text"
                        innerText = text[indexSoFar:m.start()] + m.group(2)
                        fullText = literalStart + text[indexSoFar:m.start()] + m.group(0)
                        replacement = (tag, innerText, fullText)
                        self.__codeSpanReplacements__.append(replacement)
                        newText += "\ue0ff"
                        indexSoFar = m.end()
        if mode == "text":
            newText += text[indexSoFar:]
        elif mode == "code":
            newText += tag + "`"*backtickCount + text[indexSoFar:]
        self.__val__ = newText

    def map(self, fn):
        x = MarkdownCodeSpans("")
        x.__val__ = fn(self.__val__)
        x.__codeSpanReplacements__ = self.__codeSpanReplacements__
        return x

    def extract(self):
        if self.__codeSpanReplacements__:
            # Reverse the list, so I can use pop() to get them starting from the first.
            repls = self.__codeSpanReplacements__[::-1]
            def codeSpanReviver(_):
                # Match object is the PUA character, which I can ignore.
                repl = repls.pop()
                if repl[0] == "" or repl[0].endswith("-"):
                    # Markdown code span, so massage per CommonMark rules.
                    import string
                    t = escapeHTML(repl[1]).strip(string.whitespace)
                    t = re.sub("[" + string.whitespace + "]{2,}", " ", t)
                    return "{2}<code data-opaque bs-autolink-syntax='{1}'>{0}</code>".format(t, escapeAttr(repl[2]), repl[0])
                else:
                    return "<code data-opaque data-span-tag={0} bs-autolink-syntax='{2}'>{1}</code>".format(repl[0], escapeHTML(repl[1]), escapeAttr(repl[2]))
            return re.sub("\ue0ff", codeSpanReviver, self.__val__)
        else:
            return self.__val__


def stripBOM(doc):
    if len(doc.lines) >= 1 and doc.lines[0].text[0:1] == "\ufeff":
        doc.lines[0].text = doc.lines[0].text[1:]
        warn("Your document has a BOM. There's no need for that, please re-save it without a BOM.")


# Definitions and the like

def fixManualDefTables(doc):
    # Def tables generated via datablocks are guaranteed correct,
    # but manually-written ones often don't link up the names in the first row.
    for table in findAll("table.propdef, table.descdef, table.elementdef", doc):
        if hasClass(table, "partial"):
            tag = "a"
            attr = "data-link-type"
        else:
            tag = "dfn"
            attr = "data-dfn-type"
        tag = "a" if hasClass(table, "partial") else "dfn"
        if hasClass(table, "propdef"):
            type = "property"
        elif hasClass(table, "descdef"):
            type = "descriptor"
        elif hasClass(table, "elementdef"):
            type = "element"
        cell = findAll("tr:first-child > :nth-child(2)", table)[0]
        names = [x.strip() for x in textContent(cell).split(',')]
        newContents = config.intersperse((createElement(tag, {attr:type}, name) for name in names), ", ")
        replaceContents(cell, newContents)


def canonicalizeShortcuts(doc):
    # Take all the invalid-HTML shortcuts and fix them.

    attrFixup = {
        "export":"data-export",
        "noexport":"data-noexport",
        "link-spec": "data-link-spec",
        "spec":"data-link-spec",
        "link-status": "data-link-status",
        "status":"data-link-status",
        "dfn-for":"data-dfn-for",
        "link-for":"data-link-for",
        "link-for-hint":"data-link-for-hint",
        "dfn-type":"data-dfn-type",
        "link-type":"data-link-type",
        "dfn-force": "data-dfn-force",
        "force":"data-dfn-force",
        "section":"data-section",
        "attribute-info":"data-attribute-info",
        "dict-member-info":"data-dict-member-info",
        "lt":"data-lt",
        "local-lt":"data-local-lt",
        "algorithm":"data-algorithm",
        "ignore":"data-var-ignore"
    }
    for el in findAll(",".join("[{0}]".format(attr) for attr in attrFixup.keys()), doc):
        for attr, fixedAttr in attrFixup.items():
            if el.get(attr) is not None:
                el.set(fixedAttr, el.get(attr))
                del el.attrib[attr]

    # The next two aren't in the above dict because some of the words conflict with existing attributes on some elements.
    # Instead, limit the search/transforms to the relevant elements.
    for el in findAll("dfn, h2, h3, h4, h5, h6", doc):
        for dfnType in config.dfnTypes:
            if el.get(dfnType) == "":
                del el.attrib[dfnType]
                el.set("data-dfn-type", dfnType)
                break
    for el in findAll("a", doc):
        for linkType in config.linkTypes:
            if el.get(linkType) is not None:
                del el.attrib[linkType]
                el.set("data-link-type", linkType)
                break
    for el in findAll(config.dfnElementsSelector + ", a", doc):
        if el.get("for") is None:
            continue
        if el.tag == "a":
            el.set("data-link-for", el.get('for'))
        else:
            el.set("data-dfn-for", el.get('for'))
        del el.attrib['for']


def addImplicitAlgorithms(doc):
    # If a container has an empty `algorithm` attribute,
    # but it contains only a single `<dfn>`,
    # assume that the dfn is a description of the algorithm.
    for el in findAll("[data-algorithm='']:not(h1):not(h2):not(h3):not(h4):not(h5):not(h6)", doc):
        dfns = findAll("dfn", el)
        if len(dfns) == 1:
            dfnName = config.firstLinkTextFromElement(dfns[0])
            el.set("data-algorithm", dfnName)
            dfnFor = dfns[0].get('data-dfn-for')
            if dfnFor:
                el.set("data-algorithm-for", dfnFor)
        elif len(dfns) == 0:
            die("Algorithm container has no name, and there is no <dfn> to infer one from.", el=el)
        else:
            die("Algorithm container has no name, and there are too many <dfn>s to choose which to infer a name from.", el=el)


def checkVarHygiene(doc):
    def isAlgo(el):
        return el.get("data-algorithm") is not None
    def nearestAlgo(el):
        # Find the nearest "algorithm" container,
        # either an ancestor with [algorithm] or the nearest heading with same.
        if isAlgo(el):
            return el
        ancestor = closestAncestor(el, isAlgo)
        if ancestor is not None:
            return ancestor
        for h in relevantHeadings(el):
            if isAlgo(h):
                return h
    def algoName(el):
        # Finds a uniquified algorithm name from an algo container
        algo = nearestAlgo(el)
        if algo is None:
            return None
        algoName = algo.get("data-algorithm")
        algoFor = algo.get("data-algorithm-for")
        return f"'{algoName}'" + (f" for {algoFor}" if algoFor else "")

    # Look for vars that only show up once. These are probably typos.
    singularVars = []
    varCounts = defaultdict(lambda: 0)
    for el in findAll("var:not([data-var-ignore])", doc):
        key = (foldWhitespace(textContent(el)).strip(), algoName(el))
        varCounts[key] += 1
    foldedVarCounts = defaultdict(lambda: 0)
    atLeastOneAlgo = False
    for (var,algo),count in varCounts.items():
        if algo:
            atLeastOneAlgo = True
        if count > 1:
            continue
        if var.lower() in doc.md.ignoredVars:
            continue
        key = var, algo
        foldedVarCounts[key] += 1
    varLines = []
    for (var, algo),count in foldedVarCounts.items():
        if count == 1:
            if algo:
                varLines.append(f"  '{var}', in algorithm {algo}")
            else:
                varLines.append(f"  '{var}'")
    if varLines:
        warn("The following <var>s were only used once in the document:\n{0}\nIf these are not typos, please add an ignore='' attribute to the <var>.", "\n".join(varLines))

    if atLeastOneAlgo:
        addVarClickHighlighting(doc)

    # Look for algorithms that show up twice; these are errors.
    for algo, count in Counter(algoName(el) for el in findAll("[data-algorithm]", doc)).items():
        if count > 1:
            die(f"Multiple declarations of the {algo} algorithm.")
            return


def addVarClickHighlighting(doc):
    if doc.md.slimBuildArtifact:
        return
    doc.extraStyles["style-var-click-highlighting"] = '''
    var { cursor: pointer; }
    var.selected0 { background-color: #F4D200; box-shadow: 0 0 0 2px #F4D200; }
    var.selected1 { background-color: #FF87A2; box-shadow: 0 0 0 2px #FF87A2; }
    var.selected2 { background-color: #96E885; box-shadow: 0 0 0 2px #96E885; }
    var.selected3 { background-color: #3EEED2; box-shadow: 0 0 0 2px #3EEED2; }
    var.selected4 { background-color: #EACFB6; box-shadow: 0 0 0 2px #EACFB6; }
    var.selected5 { background-color: #82DDFF; box-shadow: 0 0 0 2px #82DDFF; }
    var.selected6 { background-color: #FFBCF2; box-shadow: 0 0 0 2px #FFBCF2; }
    '''
    # Colors were chosen in Lab using https://nixsensor.com/free-color-converter/
    # D50 2deg illuminant, L in [0,100], a and b in [-128, 128]
    # 0 = lab(85,0,85)
    # 1 = lab(85,80,30)
    # 2 = lab(85,-40,40)
    # 3 = lab(85,-50,0)
    # 4 = lab(85,5,15)
    # 5 = lab(85,-10,-50)
    # 6 = lab(85,35,-15)

    # Color-choosing design.
    # Start with 1, increment as new ones get selected.
    # Specifically: find lowest-indexed color with lowest usage.
    # (Usually this'll be zero, but if you click too many vars in same algo, can repeat.)
    # If you unclick then click again on same var, it should get same color if possible.
    doc.extraScripts["script-var-click-highlighting"] = '''
    document.addEventListener("click", e=>{
        if(e.target.nodeName == "VAR") {
            highlightSameAlgoVars(e.target);
        }
    });
    {
        const indexCounts = new Map();
        const indexNames = new Map();
        function highlightSameAlgoVars(v) {
            // Find the algorithm container.
            let algoContainer = null;
            let searchEl = v;
            while(algoContainer == null && searchEl != document.body) {
                searchEl = searchEl.parentNode;
                if(searchEl.hasAttribute("data-algorithm")) {
                    algoContainer = searchEl;
                }
            }

            // Not highlighting document-global vars,
            // too likely to be unrelated.
            if(algoContainer == null) return;

            const algoName = algoContainer.getAttribute("data-algorithm");
            const varName = getVarName(v);
            const addClass = !v.classList.contains("selected");
            let highlightClass = null;
            if(addClass) {
                const index = chooseHighlightIndex(algoName, varName);
                indexCounts.get(algoName)[index] += 1;
                indexNames.set(algoName+"///"+varName, index);
                highlightClass = nameFromIndex(index);
            } else {
                const index = previousHighlightIndex(algoName, varName);
                indexCounts.get(algoName)[index] -= 1;
                highlightClass = nameFromIndex(index);
            }

            // Find all same-name vars, and toggle their class appropriately.
            for(const el of algoContainer.querySelectorAll("var")) {
                if(getVarName(el) == varName) {
                    el.classList.toggle("selected", addClass);
                    el.classList.toggle(highlightClass, addClass);
                }
            }
        }
        function getVarName(el) {
            return el.textContent.replace(/(\s|\xa0)+/, " ").trim();
        }
        function chooseHighlightIndex(algoName, varName) {
            let indexes = null;
            if(indexCounts.has(algoName)) {
                indexes = indexCounts.get(algoName);
            } else {
                // 7 classes right now
                indexes = [0,0,0,0,0,0,0];
                indexCounts.set(algoName, indexes);
            }

            // If the element was recently unclicked,
            // *and* that color is still unclaimed,
            // give it back the same color.
            const lastIndex = previousHighlightIndex(algoName, varName);
            if(indexes[lastIndex] === 0) return lastIndex;

            // Find the earliest index with the lowest count.
            const minCount = Math.min.apply(null, indexes);
            let index = null;
            for(var i = 0; i < indexes.length; i++) {
                if(indexes[i] == minCount) {
                    return i;
                }
            }
        }
        function previousHighlightIndex(algoName, varName) {
            return indexNames.get(algoName+"///"+varName);
        }
        function nameFromIndex(index) {
            return "selected" + index;
        }
    }
    '''


def fixIntraDocumentReferences(doc):
    ids = {el.get('id'):el for el in findAll("[id]", doc)}
    headingIDs = {el.get('id'):el for el in findAll("[id].heading", doc)}
    for el in findAll("a[href^='#']:not([href='#']):not(.self-link):not([data-link-type])", doc):
        targetID = urllib.parse.unquote(el.get("href")[1:])
        if el.get('data-section') is not None and targetID not in headingIDs:
            die("Couldn't find target document section {0}:\n{1}", targetID, outerHTML(el), el=el)
            continue
        elif targetID not in ids:
            die("Couldn't find target anchor {0}:\n{1}", targetID, outerHTML(el), el=el)
            continue
        if isEmpty(el):
            # TODO Allow this to respect "safe" markup (<sup>, etc) in the title
            target = ids[targetID]
            content = find(".content", target)
            if content is None:
                die("Tried to generate text for a section link, but the target isn't a heading:\n{0}", outerHTML(el), el=el)
                continue
            text = textContent(content).strip()
            if target.get('data-level') is not None:
                text = "§\u202f{1} {0}".format(text, target.get('data-level'))
            appendChild(el, text)


def fixInterDocumentReferences(doc):
    for el in findAll("[spec-section]", doc):
        spec = el.get('data-link-spec').lower()
        section = el.get('spec-section', '')
        if spec is None:
            die("Spec-section autolink doesn't have a 'spec' attribute:\n{0}", outerHTML(el), el=el)
            continue
        if section is None:
            die("Spec-section autolink doesn't have a 'spec-section' attribute:\n{0}", outerHTML(el), el=el)
            continue
        if spec in doc.refs.specs:
            # Bikeshed recognizes the spec
            fillInterDocumentReferenceFromShepherd(doc, el, spec, section)
            continue
        if not re.search(r"\d$", spec):
            # Unnumbered, check if there's a numbered variant in Shepherd
            vNames = doc.refs.vNamesFromSpecNames(spec)
            if len(vNames):
                fillInterDocumentReferenceFromShepherd(doc, el, vNames[0], section)
                if len(vNames) > 1:
                    die("Section autolink {2} attempts to link to unversioned spec name '{0}', "+
                         "but that spec is versioned as {1}. "+
                         "Please choose a versioned spec name.",
                        spec,
                        config.englishFromList("'{0}'".format(x) for x in vNames),
                        outerHTML(el),
                        el=el)
                continue
        if doc.refs.getBiblioRef(spec):
            # Bikeshed doesn't know the spec, but it's in biblio
            fillInterDocumentReferenceFromSpecref(doc, el, spec, section)
            continue
        # Unknown spec
        die("Spec-section autolink tried to link to non-existent '{0}' spec:\n{1}", spec, outerHTML(el), el=el)


def fillInterDocumentReferenceFromShepherd(doc, el, spec, section):
    specData = doc.refs.fetchHeadings(spec)
    if section in specData:
        heading = specData[section]
    else:
        die("Couldn't find section '{0}' in spec '{1}':\n{2}", section, spec, outerHTML(el), el=el)
        return
    if isinstance(heading, list):
        # Multipage spec
        if len(heading) == 1:
            # only one heading of this name, no worries
            heading = specData[heading[0]]
        else:
            # multiple headings of this id, user needs to disambiguate
            die("Multiple headings with id '{0}' for spec '{1}'. Please specify:\n{2}", section, spec, "\n".join("  [[{0}]]".format(spec + x) for x in heading), el=el)
            return
    if doc.md.status == "current":
        if "current" in heading:
            heading = heading["current"]
        else:
            heading = heading["snapshot"]
    else:
        if "snapshot" in heading:
            heading = heading["snapshot"]
        else:
            heading = heading["current"]
    el.tag = "a"
    el.set("href", heading['url'])
    if isEmpty(el):
        el.text = "{spec} §{number} {text}".format(**heading)
    removeAttr(el, 'data-link-spec', 'spec-section')

def fillInterDocumentReferenceFromSpecref(doc, el, spec, section):
    bib = doc.refs.getBiblioRef(spec)
    if isinstance(bib, biblio.StringBiblioEntry):
        die("Can't generate a cross-spec section ref for '{0}', because the biblio entry has no url.", spec, el=el)
        return
    el.tag = "a"
    el.set("href", bib.url + section)
    if isEmpty(el):
        el.text = bib.title + " §" + section[1:]
    removeAttr(el, 'data-link-spec', 'spec-section')


def fillAttributeInfoSpans(doc):
    # Auto-add <span attribute-info> to <dt><dfn> when it's an attribute or dict-member.
    for dt in findAll("dt", doc):
        if find("span[data-attribute-info]", dt) is not None:
            # Already has one, no need to do any work here
            continue
        dfn = find("dfn", dt)
        if dfn is None:
            continue
        dfnType = dfn.get("data-dfn-type")
        if dfnType == "attribute":
            attrName = "data-attribute-info"
        elif dfnType == "dict-member":
            attrName = "data-dict-member-info"
        else:
            continue
        spanFor = config.firstLinkTextFromElement(dfn)
        if spanFor is None:
            continue
        # Internal slots (denoted by [[foo]] naming scheme) don't have attribute info
        if spanFor.startswith("[["):
            continue
        if dfn.get("data-dfn-for"):
            spanFor = dfn.get("data-dfn-for") + "/" + spanFor
        # If there's whitespace after the dfn, clear it out
        if emptyText(dfn.tail, wsAllowed=True):
            dfn.tail = None
        insertAfter(dfn,
                    ", ",
                    E.span({attrName:"", "for":spanFor}))

    for el in findAll("span[data-attribute-info], span[data-dict-member-info]", doc):
        if el.get('data-attribute-info') is not None:
            refType = "attribute"
        else:
            refType = "dict-member"
        if (el.text is None or el.text.strip() == '') and len(el) == 0:
            referencedAttribute = el.get("for")
            if referencedAttribute is None or referencedAttribute == "":
                die("Missing for reference in attribute info span.", el=el)
                continue
            if "/" in referencedAttribute:
                interface, referencedAttribute = referencedAttribute.split("/")
                target = findAll('a[data-link-type={2}][data-lt="{0}"][data-link-for="{1}"]'.format(referencedAttribute, interface, refType), doc)
            else:
                target = findAll('a[data-link-type={1}][data-lt="{0}"]'.format(referencedAttribute, refType), doc)
            if len(target) == 0:
                die("Couldn't find target {1} '{0}':\n{2}", referencedAttribute, refType, outerHTML(el), el=el)
                continue
            elif len(target) > 1:
                die("Multiple potential target {1}s '{0}':\n{2}", referencedAttribute, refType, outerHTML(el), el=el)
                continue
            target = target[0]
            datatype = target.get("data-type").strip()
            default = target.get("data-default")
            decorations = []
            if target.get("data-readonly") is not None:
                decorations.append(", readonly")
            if datatype[-1] == "?":
                decorations.append(", nullable")
                datatype = datatype[:-1]
            if default is not None:
                decorations.append(", defaulting to ")
                decorations.append(E.code(default))
            if datatype[0] == "(":
                # Union type
                # TODO(Nov 2015): actually handle this properly, don't have time to think through it right now.
                appendChild(el,
                            " of type ",
                            E.code({"class":"idl-code"}, datatype),
                            *decorations)
            elif re.match(r"(\w+)<(\w+)>", datatype):
                # Sequence type
                match = re.match(r"(\w+)<(\w+)>", datatype)
                appendChild(el,
                            " of type ",
                            match.group(1),
                            "<",
                            E.a({"data-link-type":"idl-name"}, match.group(2)),
                            ">",
                            *decorations)
            else:
                # Everything else
                appendChild(el,
                            " of type ",
                            E.a({"data-link-type":"idl-name"}, datatype),
                            *decorations)


def processDfns(doc):
    dfns = findAll(config.dfnElementsSelector, doc)
    classifyDfns(doc, dfns)
    fixupIDs(doc, dfns)
    doc.refs.addLocalDfns(dfn for dfn in dfns if dfn.get('id') is not None)


def determineDfnType(dfn, inferCSS=False):
    # 1. Look at data-dfn-type
    if dfn.get('data-dfn-type'):
        return dfn.get('data-dfn-type')
    # 2. Look for a prefix on the id
    if dfn.get('id'):
        id = dfn.get('id')
        for prefix, type in config.dfnClassToType.items():
            if id.startswith(prefix):
                return type
    # 3. Look for a class or data-dfn-type on the ancestors
    for ancestor in dfn.iterancestors():
        if ancestor.get('data-dfn-type'):
            return ancestor.get('data-dfn-type')
        for cls, type in config.dfnClassToType.items():
            if hasClass(ancestor, cls):
                return type
            if hasClass(ancestor, "idl") and not hasClass(ancestor, "extract"):
                return "interface"
    # 4. Introspect on the text
    if inferCSS:
        text = textContent(dfn)
        if text[0:1] == "@":
            return "at-rule"
        elif len(dfn) == 1 and dfn[0].get('data-link-type') == "maybe" and emptyText(dfn.text) and emptyText(dfn[0].tail):
            return "value"
        elif text[0:1] == "<" and text[-1:] == ">":
            return "type"
        elif text[0:1] == ":":
            return "selector"
        elif re.match(r"^[\w-]+\(.*\)$", text) and not (dfn.get('id') or '').startswith("dom-"):
            return "function"
    # 5. Assume it's a "dfn"
    return "dfn"


def classifyDfns(doc, dfns):
    dfnTypeToPrefix = {v:k for k,v in config.dfnClassToType.items()}
    for el in dfns:
        dfnType = determineDfnType(el, inferCSS=doc.md.inferCSSDfns)
        if dfnType not in config.dfnTypes:
            die("Unknown dfn type '{0}' on:\n{1}", dfnType, outerHTML(el), el=el)
            continue
        dfnFor = treeAttr(el, "data-dfn-for")
        primaryDfnText = config.firstLinkTextFromElement(el)
        if primaryDfnText is None:
            die("Dfn has no linking text:\n{0}", outerHTML(el), el=el)
            continue
        if len(primaryDfnText) > 300:
            # Almost certainly accidentally missed the end tag
            warn("Dfn has extremely long text - did you forget the </dfn> tag?\n{0}", outerHTML(el), el=el)
        # Check for invalid fors, as it's usually some misnesting.
        if dfnFor and dfnType in config.typesNotUsingFor:
            die("'{0}' definitions don't use a 'for' attribute, but this one claims it's for '{1}' (perhaps inherited from an ancestor). This is probably a markup error.\n{2}", dfnType, dfnFor, outerHTML(el), el=el)
        # Push the dfn type down to the <dfn> itself.
        if el.get('data-dfn-type') is None:
            el.set('data-dfn-type', dfnType)
        # Push the for value too.
        if dfnFor:
            el.set('data-dfn-for', dfnFor)
        elif dfnType in config.typesUsingFor:
            die("'{0}' definitions need to specify what they're for.\nAdd a 'for' attribute to {1}, or add 'dfn-for' to an ancestor.", dfnType, outerHTML(el), el=el)
            continue
        # Some error checking
        if dfnType in config.functionishTypes:
            if not re.search(r"\(.*\)$", primaryDfnText):
                die("Function/methods must end with a () arglist in their linking text. Got '{0}'.", primaryDfnText, el=el)
                continue
            if not re.match(r"^[\w\[\]-]+\s*\(", primaryDfnText):
                die("Function/method names can only contain alphanums, underscores, dashes, or []. Got '{0}'.", primaryDfnText, el=el)
                continue
            elif el.get('data-lt') is None:
                if dfnType == "function":
                    # CSS function, define it with no args in the text
                    primaryDfnText = re.match(r"^([\w\[\]-]+)\(.*\)$", primaryDfnText).group(1) + "()"
                    el.set('data-lt', primaryDfnText)
                elif dfnType in config.idlTypes:
                    # IDL methodish construct, ask the widlparser what it should have.
                    # If the method isn't in any IDL, this tries its best to normalize it anyway.
                    names = list(doc.widl.normalized_method_names(primaryDfnText, el.get('data-dfn-for')))
                    primaryDfnText = names[0]
                    el.set('data-lt', "|".join(names))
                else:
                    die("BIKESHED ERROR: Unhandled functionish type '{0}' in classifyDfns. Please report this to Bikeshed's maintainer.", dfnType, el=el)
        # If type=argument, try to infer what it's for.
        if dfnType == "argument" and el.get('data-dfn-for') is None:
            parent = el.getparent()
            parentFor = parent.get('data-dfn-for')
            if parent.get('data-dfn-type') in config.functionishTypes and parentFor is not None:
                dfnFor = ", ".join(parentFor + "/" + name for name in doc.widl.normalized_method_names(textContent(parent), parentFor))
            elif treeAttr(el, "data-dfn-for") is None:
                die("'argument' dfns need to specify what they're for, or have it be inferrable from their parent. Got:\n{0}", outerHTML(el), el=el)
                continue
        # Automatically fill in id if necessary.
        if el.get('id') is None:
            if dfnFor:
                singleFor = config.splitForValues(dfnFor)[0]
            if dfnType in config.functionishTypes.intersection(config.idlTypes):
                id = config.simplifyText("{_for}-{id}".format(_for=singleFor, id=re.match(r"[^(]*", primaryDfnText).group(0) + "()"))
                el.set("data-alternate-id", config.simplifyText("dom-{_for}-{id}".format(_for=singleFor, id=primaryDfnText)))
                if primaryDfnText.startswith("[["):
                    # Slots get their identifying [] stripped from their ID,
                    # so gotta dedup them some other way.
                    id += "-slot"
                    el.set("data-alternate-id", "{0}-slot".format(el.get("data-alternate-id")))
            else:
                if dfnFor:
                    id = config.simplifyText("{_for}-{id}".format(_for=singleFor, id=primaryDfnText))
                else:
                    id = config.simplifyText(primaryDfnText)
            if dfnType == "dfn":
                pass
            elif dfnType == "interface":
                pass
            elif dfnType == "event":
                # Special case 'event' because it needs a different format from IDL types
                id = config.simplifyText("{type}-{id}".format(type=dfnTypeToPrefix[dfnType], _for=singleFor, id=id))
            elif dfnType == "attribute" and primaryDfnText.startswith("[["):
                # Slots get their identifying [] stripped from their ID, so gotta dedup them some other way.
                id = config.simplifyText("dom-{id}-slot".format(_for=singleFor, id=id))
            elif dfnType in config.idlTypes.intersection(config.typesUsingFor):
                id = config.simplifyText("dom-{id}".format(id=id))
            else:
                id = "{type}-{id}".format(type=dfnTypeToPrefix[dfnType], id=id)
            el.set('id', safeID(doc, id))
        # Set lt if it's not set,
        # and doing so won't mess with anything else.
        if el.get('data-lt') is None and "|" not in primaryDfnText:
            el.set('data-lt', primaryDfnText)
        # Push export/noexport down to the definition
        if el.get('data-export') is None and el.get('data-noexport') is None:
            attr,_ = closestAttr(el, "data-export", "data-noexport")
            if attr is not None:
                el.set(attr, "")
            else:
                if dfnType == "dfn":
                    el.set('data-noexport', 'by-default')
                else:
                    el.set('data-export', 'by-default')
        # If it's an code-ish type such as IDL,
        # and doesn't already have a sole <code> child,
        # wrap the contents in a <code>.
        if config.linkTypeIn(dfnType, "codelike"):
            child = hasOnlyChild(el)
            if child is not None and child.tag == "code":
                continue
            if el.tag in ("h1", "h2", "h3", "h4", "h5", "h6"):
                # Don't wrap headings, it looks bad.
                continue
            wrapContents(el, E.code())


def determineLinkType(el):
    # 1. Look at data-link-type
    linkType = treeAttr(el, 'data-link-type')
    if linkType:
        if linkType in config.linkTypes:
            return linkType
        die("Unknown link type '{0}' on:\n{1}", linkType, outerHTML(el), el=el)
        return "unknown-type"
    # 2. Introspect on the text
    text = textContent(el)
    if config.typeRe["at-rule"].match(text):
        return "at-rule"
    elif config.typeRe["type"].match(text):
        return "type"
    elif config.typeRe["selector"].match(text):
        return "selector"
    elif config.typeRe["function"].match(text):
        return "functionish"
    else:
        return "dfn"


def determineLinkText(el):
    linkType = el.get('data-link-type')
    contents = textContent(el)
    if el.get('data-lt'):
        linkText = el.get('data-lt')
    elif config.linkTypeIn(linkType, "function") and re.match(r"^[\w-]+\(.*\)$", contents):
        # Remove arguments from CSS function autolinks,
        # as they should always be defined argument-less
        # (and this allows filled-in examples to still autolink).
        linkText = re.match(r"^([\w-]+)\(.*\)$", contents).group(1) + "()"
    else:
        linkText = contents
    linkText = foldWhitespace(linkText)
    if len(linkText) == 0:
        die("Autolink {0} has no linktext.", outerHTML(el), el=el)
    return linkText


def classifyLink(el):
    linkType = determineLinkType(el)
    el.set('data-link-type', linkType)

    linkText = determineLinkText(el)
    el.set('data-lt', linkText)

    for attr in ["data-link-status", "data-link-for", "data-link-spec", "data-link-for-hint"]:
        val = treeAttr(el, attr)
        if val is not None:
            el.set(attr, val)
    return el

# Additional Processing


def processBiblioLinks(doc):
    biblioLinks = findAll("a[data-link-type='biblio']", doc)
    for el in biblioLinks:
        biblioType = el.get('data-biblio-type')
        if biblioType == "normative":
            storage = doc.normativeRefs
        elif biblioType == "informative":
            storage = doc.informativeRefs
        else:
            die("Unknown data-biblio-type value '{0}' on {1}. Only 'normative' and 'informative' allowed.", biblioType, outerHTML(el), el=el)
            continue

        linkText = determineLinkText(el)
        if linkText[0] == "[" and linkText[-1] == "]":
            linkText = linkText[1:-1]

        refStatus = treeAttr(el, "data-biblio-status") or doc.md.defaultRefStatus

        okayToFail = el.get('data-okay-to-fail') is not None

        ref = doc.refs.getBiblioRef(linkText, status=refStatus, generateFakeRef=okayToFail, quiet=okayToFail, el=el)
        if not ref:
            if not okayToFail:
                closeBiblios = biblio.findCloseBiblios(doc.refs.biblioKeys, linkText)
                die("Couldn't find '{0}' in bibliography data. Did you mean:\n{1}", linkText, '\n'.join("  " + b for b in closeBiblios), el=el)
            el.tag = "span"
            continue

        # Need to register that I have a preferred way to refer to this biblio,
        # in case aliases show up - they all need to use my preferred key!
        if ref.originalLinkText:
            # Okay, so this particular ref has been reffed before...
            if linkText == ref.linkText:
                # Whew, and with the same name I'm using now. Ship it.
                pass
            else:
                # Oh no! I'm using two different names to refer to the same biblio!
                die("The biblio refs [[{0}]] and [[{1}]] are both aliases of the same base reference [[{2}]]. Please choose one name and use it consistently.", linkText, ref.linkText, ref.originalLinkText, el=el)
                # I can keep going, tho - no need to skip this ref
        else:
            # This is the first time I've reffed this particular biblio.
            # Register this as the preferred name...
            doc.refs.preferredBiblioNames[ref.linkText] = linkText
            # Use it on the current ref. Future ones will use the preferred name automatically.
            ref.linkText = linkText

        id = config.simplifyText(ref.linkText)
        el.set('href', '#biblio-' + id)
        storage[ref.linkText] = ref


def verifyUsageOfAllLocalBiblios(doc):
    '''
    Verifies that all the locally-declared biblios
    (those written inline in a <pre class=biblio> block,
    and thus given order=1)
    were used in the spec,
    so you can remove entries when they're no longer necessary.
    '''
    usedBiblioKeys = set(x.lower() for x in list(doc.normativeRefs.keys()) + list(doc.informativeRefs.keys()))
    localBiblios = [b["linkText"].lower() for bs in doc.refs.biblios.values() for b in bs if b['order'] == 1]
    unusedBiblioKeys = []
    for b in localBiblios:
        if b not in usedBiblioKeys:
            unusedBiblioKeys.append(b)
    if unusedBiblioKeys:
        warn("The following locally-defined biblio entries are unused and can be removed:\n{0}",
             "\n".join("  * {0}".format(b) for b in unusedBiblioKeys))


def processAutolinks(doc):
    # An <a> without an href is an autolink.
    # <i> is a legacy syntax for term autolinks. If it links up, we change it into an <a>.
    # We exclude bibliographical links, as those are processed in `processBiblioLinks`.
    query = "a:not([href]):not([data-link-type='biblio'])"
    if doc.md.useIAutolinks:
        warn("Use <i> Autolinks is deprecated and will be removed. Please switch to using <a> elements.")
        query += ", i"
    autolinks = findAll(query, doc)
    for el in autolinks:
        # Explicitly empty linking text indicates this shouldn't be an autolink.
        if el.get('data-lt') == '':
            continue

        classifyLink(el)
        linkType = el.get('data-link-type')
        linkText = el.get('data-lt')

        # Properties and descriptors are often written like 'foo-*'. Just ignore these.
        if linkType in ("property", "descriptor", "propdesc") and "*" in linkText:
            continue

        # Not super clear why I think links will specify multiple for values,
        # or why it's okay to just use the first one in that case.
        linkFor = config.splitForValues(el.get('data-link-for'))
        if linkFor:
            linkFor = linkFor[0]

        # Status used to use ED/TR, so convert those if they appear,
        # and verify
        status = el.get('data-link-status')
        if status == "ED":
            status = "current"
        elif status == "TR":
            status = "snapshot"
        elif status in config.linkStatuses or status is None:
            pass
        else:
            die("Unknown link status '{0}' on {1}", status, outerHTML(el))
            continue

        ref = doc.refs.getRef(linkType, linkText,
                              spec=el.get('data-link-spec'),
                              status=status,
                              linkFor=linkFor,
                              linkForHint=el.get('data-link-for-hint'),
                              explicitFor=doc.md.assumeExplicitFor,
                              el=el,
                              error=(linkText.lower() not in doc.md.ignoredTerms))
        # Capture the reference (and ensure we add a biblio entry) if it
        # points to an external specification. We check the spec name here
        # rather than checking `status == "local"`, as "local" refs include
        # those defined in `<pre class="anchor">` datablocks, which we do
        # want to capture here.
        if ref and ref.spec and doc.refs.spec and ref.spec.lower() != doc.refs.spec.lower():
            spec = ref.spec.lower()
            key = ref.for_[0] if ref.for_ else ""
            if isNormative(el, doc):
                biblioStorage = doc.normativeRefs
            else:
                biblioStorage = doc.informativeRefs
            biblioRef = doc.refs.getBiblioRef(ref.spec, status=doc.md.defaultRefStatus, generateFakeRef=True, quiet=True)
            if biblioRef:
                biblioStorage[biblioRef.linkText] = biblioRef
                spec = biblioRef.linkText.lower()
            doc.externalRefsUsed[spec][ref.text][key] = ref

        if ref:
            el.set('href', ref.url)
            el.tag = "a"
            decorateAutolink(doc, el, linkType=linkType, linkText=linkText, ref=ref)
        else:
            if linkType == "maybe":
                el.tag = "css"
                if el.get("data-link-type"):
                    del el.attrib["data-link-type"]
                if el.get("data-lt"):
                    del el.attrib["data-lt"]
    dedupIDs(doc)


def decorateAutolink(doc, el, linkType, linkText, ref):
    # Add additional effects to autolinks.
    if doc.md.slimBuildArtifact:
        return

    # Put an ID on every reference, so I can link to references to a term.
    if el.get('id') is None:
        _,_,id = ref.url.partition("#")
        if id:
            el.set('id', "ref-for-{0}".format(id))
            el.set('data-silently-dedup', '')

    # Get all the values that the type expands to, add it as a title.
    if linkType == "type":
        titleText = None
        if linkText in doc.typeExpansions:
            titleText = doc.typeExpansions[linkText]
        else:
            refs = doc.refs.queryAllRefs(linkFor=linkText, ignoreObsoletes=True)
            texts = sorted({ref.text for ref in refs})
            if refs:
                titleText = "Expands to: " + ' | '.join(texts)
                doc.typeExpansions[linkText] = titleText
        if titleText:
            el.set('title', titleText)


def removeMultipleLinks(doc):
    # If there are multiple autolinks to the same thing in a paragraph,
    # only keep the first.
    if not doc.md.removeMultipleLinks:
        return
    paras = defaultdict(lambda:defaultdict(list))
    for el in findAll("a[data-link-type]", doc):
        if hasAncestor(el, lambda x:x.tag in ["pre", "xmp"]):
            # Don't strip out repeated links from opaque elements
            continue
        paras[parentElement(el)][el.get("href")].append(el)
    for linkGroups in paras.values():
        for href,links in linkGroups.items():
            if len(links) > 1:
                for el in links[1:]:
                    el.tag = "span"
                    removeAttr(el, "href", "data-link-type")

def processIssuesAndExamples(doc):
    # Add an auto-genned and stable-against-changes-elsewhere id to all issues and
    # examples, and link to remote issues if possible:
    for el in findAll(".issue:not([id])", doc):
        el.set('id', safeID(doc, "issue-" + hashContents(el)))
        remoteIssueID = el.get('data-remote-issue-id')
        if remoteIssueID:
            del el.attrib['data-remote-issue-id']
            # Eventually need to support a way to trigger other repo url structures,
            # but defaulting to GH is fine for now.
            githubMatch = re.match(r"\s*([\w-]+)/([\w-]+)#(\d+)\s*$", remoteIssueID)
            numberMatch = re.match(r"\s*(\d+)\s*$", remoteIssueID)
            remoteIssueURL = None
            if githubMatch:
                remoteIssueURL = "https://github.com/{0}/{1}/issues/{2}".format(*githubMatch.groups())
                if doc.md.inlineGithubIssues:
                    el.set("data-inline-github", "{0} {1} {2}".format(*githubMatch.groups()))
            elif numberMatch and doc.md.repository.type == "github":
                remoteIssueURL = doc.md.repository.formatIssueUrl(numberMatch.group(1))
                if doc.md.inlineGithubIssues:
                    el.set("data-inline-github", "{0} {1} {2}".format(doc.md.repository.user, doc.md.repository.repo, numberMatch.group(1)))
            elif doc.md.issueTrackerTemplate:
                remoteIssueURL = doc.md.issueTrackerTemplate.format(remoteIssueID)
            if remoteIssueURL:
                appendChild(el, " ", E.a({"href": remoteIssueURL}, "<" + remoteIssueURL + ">"))
    for el in findAll(".example:not([id])", doc):
        el.set('id', safeID(doc, "example-" + hashContents(el)))
    fixupIDs(doc, findAll(".issue, .example", doc))


def addSelfLinks(doc):
    if doc.md.slimBuildArtifact:
        return
    def makeSelfLink(el):
        return E.a({"href": "#" + escapeUrlFrag(el.get('id', '')), "class":"self-link"})

    dfnElements = findAll(config.dfnElementsSelector, doc)

    foundFirstNumberedSection = False
    for el in findAll("h2, h3, h4, h5, h6", doc):
        foundFirstNumberedSection = foundFirstNumberedSection or (el.get('data-level') is not None)
        if el in dfnElements:
            # It'll get a self-link or dfn-panel later.
            continue
        if foundFirstNumberedSection:
            appendChild(el, makeSelfLink(el))
    for el in findAll(".issue[id], .example[id], .note[id], li[id], dt[id]", doc):
        if list(el.iterancestors("figure")):
            # Skipping - element is inside a figure and is part of an example.
            continue
        if el.get("data-no-self-link") is not None:
            continue
        if el.tag == "details":
            summary = find("summary", el)
            if summary is not None:
                insertAfter(summary, makeSelfLink(el))
                continue
        prependChild(el, makeSelfLink(el))
    if doc.md.useDfnPanels:
        dfnpanels.addDfnPanels(doc, dfnElements)
    else:
        for el in dfnElements:
            if list(el.iterancestors("a")):
                warn("Found <a> ancestor, skipping self-link. Swap <dfn>/<a> order?\n  {0}", outerHTML(el), el=el)
                continue
            appendChild(el, makeSelfLink(el))


def cleanupHTML(doc):
    # Cleanup done immediately before serialization.

    head = None
    inBody = False
    strayHeadEls = []
    styleScoped = []
    nestedLists = []
    flattenEls = []
    for el in doc.document.iter():
        if head is None and el.tag == "head":
            head = el
            continue
        if el.tag == "body":
            inBody = True

        # Move any stray <link>, <meta>, or <style> into the <head>.
        if inBody and el.tag in ["link", "meta", "style"]:
            strayHeadEls.append(el)

        if el.tag == "style" and el.get("scoped") is not None:
            die("<style scoped> is no longer part of HTML. Ensure your styles can apply document-globally and remove the scoped attribute.", el=el)
            styleScoped.append(el)

        # Convert the technically-invalid <nobr> element to an appropriate <span>
        if el.tag == "nobr":
            el.tag = "span"
            el.set("style", el.get('style', '') + ";white-space:nowrap")

        # And convert <xmp> to <pre>
        if el.tag == "xmp":
            el.tag = "pre"

        # If we accidentally recognized an autolink shortcut in SVG, kill it.
        if el.tag == "{http://www.w3.org/2000/svg}a" and el.get("data-link-type") is not None:
            removeAttr(el, "data-link-type")
            el.tag = "{http://www.w3.org/2000/svg}tspan"

        # Add .algorithm to [algorithm] elements, for styling
        if el.get("data-algorithm") is not None and not hasClass(el, "algorithm"):
            addClass(el, "algorithm")

        # Allow MD-generated lists to be surrounded by HTML list containers,
        # so you can add classes/etc without an extraneous wrapper.
        if el.tag in ["ol", "ul", "dl"]:
            onlyChild = hasOnlyChild(el)
            if onlyChild is not None and el.tag == onlyChild.tag and el.get("data-md") is None and onlyChild.get("data-md") is not None:
                # The md-generated list container is featureless,
                # so we can just throw it away and move its children into its parent.
                nestedLists.append(onlyChild)
            else:
                # Remove any lingering data-md attributes on lists that weren't using this container replacement thing.
                removeAttr(el, "data-md")

        # Mark pre.idl blocks as .def, for styling
        if el.tag == "pre" and hasClass(el, "idl") and not hasClass(el, "def"):
            addClass(el, "def")

        # Tag classes on wide types of dfns/links
        if el.tag in config.dfnElements:
            if el.get("data-dfn-type") in config.idlTypes:
                addClass(el, "idl-code")
            if el.get("data-dfn-type") in config.maybeTypes.union(config.linkTypeToDfnType['propdesc']):
                if not hasAncestor(el, lambda x:x.tag=="pre"):
                    addClass(el, "css")
        if el.tag == "a":
            if el.get("data-link-type") in config.idlTypes:
                addClass(el, "idl-code")
            if el.get("data-link-type") in config.maybeTypes.union(config.linkTypeToDfnType['propdesc']):
                if not hasAncestor(el, lambda x:x.tag=="pre"):
                    addClass(el, "css")

        # Remove duplicate linking texts.
        if el.tag in config.anchorishElements and el.get("data-lt") is not None and el.get("data-lt") == textContent(el, exact=True):
            removeAttr(el, "data-lt")

        # Transform the <css> fake tag into markup.
        # (Used when the ''foo'' shorthand doesn't work.)
        if el.tag == "css":
            el.tag = "span"
            addClass(el, "css")

        # Transform the <assert> fake tag into a span with a unique ID based on its contents.
        # This is just used to tag arbitrary sections with an ID so you can point tests at it.
        # (And the ID will be guaranteed stable across publications, but guaranteed to change when the text changes.)
        if el.tag == "assert":
            el.tag = "span"
            el.set("id", safeID(doc, "assert-" + hashContents(el)))

        # Add ARIA role of "note" to class="note" elements
        if el.tag in ["div", "p"] and hasClass(el, doc.md.noteClass):
            el.set("role", "note")

        # Look for nested <a> elements, and warn about them.
        if el.tag == "a" and hasAncestor(el, lambda x:x.tag=="a"):
            warn("The following (probably auto-generated) link is illegally nested in another link:\n{0}", outerHTML(el), el=el)

        # If the <h1> contains only capital letters, add a class=allcaps for styling hook
        if el.tag == "h1":
            for letter in textContent(el):
                if letter.isalpha() and letter.islower():
                    break
            else:
                addClass(el, "allcaps")

        # If a markdown-generated <dt> contains only a single paragraph,
        # remove that paragraph so it just contains naked text.
        if el.tag == "dt" and el.get("data-md") is not None:
            child = hasOnlyChild(el)
            if child is not None and child.tag == "p" and emptyText(el.text) and emptyText(child.tail):
                flattenEls.append(el)

        # Remove a bunch of attributes
        if el.get("data-attribute-info") is not None or el.get("data-dict-member-info") is not None:
            removeAttr(el, 'data-attribute-info', 'data-dict-member-info', 'for')
        if el.tag in ["a", "span"]:
            removeAttr(el, 'data-link-for', 'data-link-for-hint', 'data-link-status', 'data-link-spec', 'data-section', 'data-biblio-type', 'data-biblio-status', 'data-okay-to-fail', 'data-lt')
        if el.tag != "a":
            removeAttr(el, 'data-link-for', 'data-link-type')
        if el.tag not in config.dfnElements:
            removeAttr(el, 'data-dfn-for', 'data-dfn-type', 'data-export', 'data-noexport')
        if el.tag == "var":
            removeAttr(el, 'data-var-ignore')
        removeAttr(el, 'bs-autolink-syntax', 'data-alternate-id', 'highlight', 'nohighlight', 'line-numbers', 'data-opaque', 'data-no-self-link', "line-number", "caniuse", "data-silently-dedup", "nocrossorigin")

        # Remove the internal-use-only detail of whether export/noexport is manual or default
        if el.get("data-export"):
            el.set("data-export", "")
        if el.get("data-noexport"):
            el.set("data-noexport", "")

        if doc.md.slimBuildArtifact:
            # Remove *all* data- attributes.
            for attrName in el.attrib:
                if attrName.startswith("data-"):
                    removeAttr(el, attrName)
    for el in strayHeadEls:
        head.append(el)
    for el in styleScoped:
        parent = parentElement(el)
        prependChild(parent, el)
    for el in nestedLists:
        children = childNodes(el, clear=True)
        parent = parentElement(el)
        clearContents(parent)
        appendChild(parent, *children)
    for el in flattenEls:
        moveContents(fromEl=el[0], toEl=el)


def finalHackyCleanup(text):
    # For hacky last-minute string-based cleanups of the rendered html.

    return text


def hackyLineNumbers(lines):
    # Hackily adds line-number information to each thing that looks like an open tag.
    # This is just regex text-munging, so potentially dangerous!
    for line in lines:
        line.text = re.sub(r"(^|[^<])(<[\w-]+)([ >])", r"\1\2 line-number={0}\3".format(line.i), line.text)
    return lines


def correctH1(doc):
    # If you provided an <h1> manually, use that element rather than whatever the boilerplate contains.
    h1s = [h1 for h1 in findAll("h1", doc) if isNormative(h1, doc)]
    if len(h1s) == 2:
        replaceNode(h1s[0], h1s[1])


def formatElementdefTables(doc):
    for table in findAll("table.elementdef", doc):
        elements = findAll("tr:first-child dfn", table)
        elementsFor = ' '.join(textContent(x) for x in elements)
        for el in findAll("a[data-element-attr-group]", table):
            groupName = textContent(el).strip()
            groupAttrs = sorted(doc.refs.queryAllRefs(linkType="element-attr", linkFor=groupName), key=lambda x:x.text)
            if len(groupAttrs) == 0:
                die("The element-attr group '{0}' doesn't have any attributes defined for it.", groupName, el=el)
                continue
            el.tag = "details"
            clearContents(el)
            removeAttr(el, "data-element-attr-group", "data-dfn-type")
            ul = appendChild(el,
                             E.summary(
                                 E.a({"data-link-type":"dfn"}, groupName)),
                             E.ul())
            for ref in groupAttrs:
                id = "element-attrdef-" + config.simplifyText(textContent(elements[0])) + "-" + ref.text
                appendChild(ul,
                            E.li(
                                E.dfn({"id": safeID(doc, id), "for":elementsFor, "data-dfn-type":"element-attr"},
                                      E.a({"data-link-type":"element-attr", "for":groupName},
                                          ref.text.strip()))))


def formatArgumentdefTables(doc):
    for table in findAll("table.argumentdef", doc):
        forMethod = doc.widl.normalized_method_names(table.get("data-dfn-for"))
        method = doc.widl.find(table.get("data-dfn-for"))
        if not method:
            die("Can't find method '{0}'.", forMethod, el=table)
            continue
        for tr in findAll("tbody > tr", table):
            tds = findAll("td", tr)
            argName = textContent(tds[0]).strip()
            arg = method.find_argument(argName)
            if arg:
                appendChild(tds[1], str(arg.type))
                if str(arg.type).strip().endswith("?"):
                    appendChild(tds[2],
                                E.span({"class":"yes"}, "✔"))
                else:
                    appendChild(tds[2],
                                E.span({"class":"no"}, "✘"))
                if arg.optional:
                    appendChild(tds[3],
                                E.span({"class":"yes"}, "✔"))
                else:
                    appendChild(tds[3],
                                E.span({"class":"no"}, "✘"))
            else:
                die(f"Can't find the '{argName}' argument of method '{method.full_name}' in the argumentdef block.", el=table)
                continue


def inlineRemoteIssues(doc):
    # Finds properly-marked-up "remote issues",
    # and inlines their contents into the issue.

    # Right now, only github inline issues are supported.
    # More can be supported when someone cares.

    # Collect all the inline issues in the document
    inlineIssues = []
    GitHubIssue = namedtuple('GitHubIssue', ['user', 'repo', 'num', 'el'])
    for el in findAll("[data-inline-github]", doc):
        inlineIssues.append(GitHubIssue(*el.get('data-inline-github').split(), el=el))
        removeAttr(el, "data-inline-github")
    if not inlineIssues:
        return

    import requests

    logging.captureWarnings(True)

    responses = json.loads(doc.dataFile.fetch("github-issues.json", str=True))
    for i,issue in enumerate(inlineIssues):
        issueUserRepo = "{0}/{1}".format(*issue)
        key = "{0}/{1}".format(issueUserRepo, issue.num)
        href = "https://github.{0}/{1}/issues/{2}".format(doc.md.repository.ns, issueUserRepo, issue.num)
        url = "{0}/repos/{1}/issues/{2}".format(doc.md.repository.api, issueUserRepo, issue.num)
        say("Fetching issue {:-3d}/{:d}: {:s}".format(i+1, len(inlineIssues), key))

        # Fetch the issues
        headers = {"Accept": "application/vnd.github.v3.html+json"}
        if doc.token is not None:
            headers["Authorization"] = "token " + doc.token
        if key in responses:
            # Have a cached response, see if it changed
            headers["If-None-Match"] = responses[key]["ETag"]

        res = None
        try:
            res = requests.get(url, headers=headers)
        except requests.exceptions.ConnectionError:
            # Offline or something, recover if possible
            if key in responses:
                data = responses[key]
            else:
                warn("Connection error fetching issue #{0}", issue.num)
                continue
        if res is None:
            # Already handled in the except block
            pass
        elif res.status_code == 304:
            # Unchanged, I can use the cache
            data = responses[key]
        elif res.status_code == 200:
            # Fresh data, prep it for storage
            data = res.json()
            data["ETag"] = res.headers["ETag"]
        elif res.status_code == 401:
            error = res.json()
            if error["message"] == "Bad credentials":
                die("'{0}' is not a valid GitHub OAuth token. See https://github.com/settings/tokens", doc.token)
            else:
                die("401 error when fetching GitHub Issues:\n{0}", config.printjson(error))
            continue
        elif res.status_code == 403:
            error = res.json()
            if error["message"].startswith("API rate limit exceeded"):
                die("GitHub Issues API rate limit exceeded. Get an OAuth token from https://github.com/settings/tokens to increase your limit, or just wait an hour for your limit to refresh; Bikeshed has cached all the issues so far and will resume from where it left off.")
            else:
                die("403 error when fetching GitHub Issues:\n{0}", config.printjson(error))
            continue
        elif res.status_code >= 400:
            try:
                error = config.printjson(res.json())
            except:
                error = "First 100 characters of error:\n" + res.text[0:100]
            die("{0} error when fetching GitHub Issues:\n{1}", res.status_code, error)
            continue
        responses[key] = data
        # Put the issue data into the DOM
        el = issue.el
        data = responses[key]
        clearContents(el)
        if doc.md.inlineGithubIssues == 'title':
            appendChild(el,
                    E.a({"href":href, "class":"marker", "style":"text-transform:none"}, key),
                    E.a({"href":href}, data['title']))
            addClass(el, "no-marker")
        else:
            appendChild(el,
                    E.a({"href":href, "class":"marker"},
                    "Issue #{0} on GitHub: “{1}”".format(data['number'], data['title'])),
                    *parseHTML(data['body_html']))
            addClass(el, "no-marker")
        if el.tag == "p":
            el.tag = "div"
    # Save the cache for later
    try:
        with io.open(config.scriptPath("spec-data", "github-issues.json"), 'w', encoding="utf-8") as f:
            f.write(json.dumps(responses, ensure_ascii=False, indent=2, sort_keys=True))
    except Exception as e:
        warn("Couldn't save GitHub Issues cache to disk.\n{0}", e)
    return


def addNoteHeaders(doc):
    # Finds <foo heading="bar"> and turns it into a marker-heading
    for el in findAll("[heading]", doc):
        addClass(el, "no-marker")
        if hasClass(el, "note"):
            preText = "NOTE: "
        elif hasClass(el, "issue"):
            preText = "ISSUE: "
        elif hasClass(el, "example"):
            preText = "EXAMPLE: "
        else:
            preText = ""
        prependChild(el,
                     E.div({"class":"marker"}, preText, *parseHTML(el.get('heading'))))
        removeAttr(el, "heading")


def locateFillContainers(doc):
    fillContainers = defaultdict(list)
    for el in findAll("[data-fill-with]", doc):
        fillContainers[el.get("data-fill-with")].append(el)
    return fillContainers


def forceCrossorigin(doc):
    if not doc.md.forceCrossorigin:
        return
    for el in findAll("link, script[src], audio, video, img", doc):
        if el.get("crossorigin") is not None or treeAttr(el, "nocrossorigin") is not None:
            continue
        el.set("crossorigin", "")



# -*- coding: utf-8 -*-

import copy
import os
import re
import subprocess
from collections import defaultdict, OrderedDict
from datetime import datetime
from . import config
from . import dfnpanels
from .refs import utils as refUtils
from .DefaultOrderedDict import DefaultOrderedDict
from .html import *
from .messages import *


def addBikeshedVersion(doc):
    # Adds a <meta> containing the current Bikeshed semver.
    if "generator" not in doc.md.boilerplate:
        return
    try:
        # Check that we're in the bikeshed repo
        origin = subprocess.check_output(
            "git remote -v",
            cwd=config.scriptPath(),
            stderr=subprocess.DEVNULL,
            shell=True).decode(encoding="utf-8")
        if "bikeshed" not in origin:
            # In a repo, but not bikeshed's;
            # probably pip-installed into an environment's repo or something.
            raise Exception()
        # Otherwise, success, this is a -e install,
        # so we're in Bikeshed's repo.
        bikeshedVersion = subprocess.check_output(
            r"git log -1 --format='Bikeshed version %h, updated %cd'",
            cwd=config.scriptPath(),
            stderr=subprocess.DEVNULL,
            shell=True).decode(encoding="utf-8").strip()
    except Exception as e:
        # Not in Bikeshed's repo, so instead grab from the datafile.
        bikeshedVersion = doc.dataFile.fetch("bikeshed-version.txt", type="readonly", str=True).strip()
    appendChild(doc.head,
                E.meta({"name": "generator", "content": bikeshedVersion}))


def addCanonicalURL(doc):
    # Adds a <link rel=canonical> to the configured canonical url
    if doc.md.canonicalURL:
        appendChild(doc.head,
                    E.link({"rel": "canonical", "href": doc.md.canonicalURL}))

def addFavicon(doc):
    # Adds a <link rel=icon> to the configured favicon url
    if doc.md.favicon:
        appendChild(doc.head,
                    E.link({"rel": "icon", "href": doc.md.favicon}))


def addSpecVersion(doc):
    # Adds a <meta> with the current spec revision, if one was detected
    if "document-revision" not in doc.md.boilerplate:
        return

    if not doc.inputSource.hasDirectory():
        return

    revision = None
    source_dir = doc.inputSource.directory()
    try:
        # Check for a Git repo
        with open(os.devnull, "wb") as fnull:
            revision = subprocess.check_output("git rev-parse HEAD", stderr=fnull, shell=True, cwd=source_dir).decode(encoding="utf-8").strip()
    except subprocess.CalledProcessError:
        try:
            # Check for an Hg repo
            with open(os.devnull, "wb") as fnull:
                revision = subprocess.check_output("hg parent --temp='{node}'", stderr=fnull, shell=True, cwd=source_dir).decode(encoding="utf-8").strip()
        except:
            pass
    if revision:
        appendChild(doc.head,
                    E.meta({"name":"document-revision", "content":revision}))


def addHeaderFooter(doc):
    header = config.retrieveBoilerplateFile(doc, 'header') if "header" in doc.md.boilerplate else ""
    footer = config.retrieveBoilerplateFile(doc, 'footer') if "footer" in doc.md.boilerplate else ""

    doc.html = '\n'.join([header, doc.html, footer])


def fillWith(tag, newElements, doc):
    for el in doc.fillContainers[tag]:
        replaceContents(el, newElements)


def getFillContainer(tag, doc, default=False):
    '''
    Gets the element that should be filled with the stuff corresponding to tag.
    If it returns None, don't generate the section.

    If default=True,
    indicates that this is a "default on" section,
    and will be appended to <body> unless explicitly suppressed.
    Otherwise,
    it'll only be appended if explicitly requested with a data-fill-with attribute.
    '''

    # If you've explicitly suppressed that section, don't do anything
    if tag not in doc.md.boilerplate:
        return None

    # If a fill-with is found, fill that
    if doc.fillContainers[tag]:
        return doc.fillContainers[tag][0]

    # Otherwise, append to the end of the document,
    # unless you're in the byos group
    if doc.md.group == "byos":
        return None
    if default:
        return doc.body


def addLogo(doc):
    html = config.retrieveBoilerplateFile(doc, 'logo')
    html = doc.fixText(html)
    fillWith('logo', parseHTML(html), doc=doc)


def addCopyright(doc):
    html = config.retrieveBoilerplateFile(doc, 'copyright')
    html = doc.fixText(html)
    fillWith('copyright', parseHTML(html), doc=doc)


def addAbstract(doc):
    if not doc.md.noAbstract:
        html = config.retrieveBoilerplateFile(doc, 'abstract')
        html = doc.fixText(html)
        fillWith('abstract', parseHTML(html), doc=doc)
    else:
        container = getFillContainer("abstract", doc, default=False)
        if container is not None:
            removeNode(container)


def addStatusSection(doc):
    html = config.retrieveBoilerplateFile(doc, 'status')
    html = doc.fixText(html)
    fillWith('status', parseHTML(html), doc=doc)


def addExpiryNotice(doc):
    if doc.md.expires is None:
        return
    if doc.md.date >= doc.md.expires or datetime.utcnow().date() >= doc.md.expires:
        boilerplate = "warning-expired"
    else:
        boilerplate = "warning-expires"
        doc.extraScripts['script-expires'] = expiryScript
    html = config.retrieveBoilerplateFile(doc, boilerplate)
    html = doc.fixText(html)
    fillWith('warning', parseHTML(html), doc=doc)
    addClass(doc.body, boilerplate)

expiryScript = '''
const warning = document.querySelector('#expiry-notice');
const expiresOn = warning.dataset.expires;
const today = new Date().toISOString();
if(expires < today) {
    warning.setAttribute("open", "");
    for(const swap of warning.querySelectorAll("[data-after-expiry]")) {
        swap.textContent = swap.dataset.afterExpiry;
    }
}
'''


def addObsoletionNotice(doc):
    if doc.md.warning:
        html = config.retrieveBoilerplateFile(doc, doc.md.warning[0])
        html = doc.fixText(html)
        fillWith('warning', parseHTML(html), doc=doc)


def addAtRisk(doc):
    if len(doc.md.atRisk) == 0:
        return
    html = "<p>The following features are at-risk, and may be dropped during the CR period:\n<ul>"
    for feature in doc.md.atRisk:
        html += "<li>" + doc.fixText(feature)
    html += "</ul><p>“At-risk” is a W3C Process term-of-art, and does not necessarily imply that the feature is in danger of being dropped or delayed. It means that the WG believes the feature may have difficulty being interoperably implemented in a timely manner, and marking it as such allows the WG to drop the feature if necessary when transitioning to the Proposed Rec stage, without having to publish a new Candidate Rec without the feature first."
    fillWith('at-risk', parseHTML(html), doc=doc)


def addStyles(doc):
    el = getFillContainer('stylesheet', doc)
    if el is not None:
        el.text = config.retrieveBoilerplateFile(doc, 'stylesheet')


def addCustomBoilerplate(doc):
    for el in findAll('[boilerplate]', doc):
        tag = el.get('boilerplate')
        if doc.fillContainers[tag]:
            replaceContents(doc.fillContainers[tag][0], el)
            removeNode(el)


def removeUnwantedBoilerplate(doc):
    for tag,els in doc.fillContainers.items():
        if tag not in doc.md.boilerplate:
            for el in els:
                removeNode(el)


def addAnnotations(doc):
    if doc.md.vshortname in doc.testSuites:
        html = config.retrieveBoilerplateFile(doc, 'annotations')
        html = doc.fixText(html)
        appendContents(find("head", doc), parseHTML(html))


def addBikeshedBoilerplate(doc):
    for k,v in sorted(doc.extraStyles.items()):
        if k not in doc.md.boilerplate:
            continue
        container = getFillContainer(k, doc)
        if container is None:
            container = getFillContainer("bs-styles", doc, default=True)
        if container is not None:
            appendChild(container,
                        E.style("/* {0} */\n".format(k) + v))
    for k,v in sorted(doc.extraScripts.items()):
        if k not in doc.md.boilerplate:
            continue
        container = getFillContainer(k, doc)
        if container is None:
            container = getFillContainer("bs-scripts", doc, default=True)
        if container is not None:
            appendChild(container,
                        E.script("/* {0} */\n".format(k) + v))


def addIndexSection(doc):
    if len(findAll(config.dfnElementsSelector, doc)) == 0 and len(list(doc.externalRefsUsed.keys())) == 0:
        return
    container = getFillContainer('index', doc=doc, default=True)
    if container is None:
        return
    appendChild(container,
                E.h2({"class":"no-num no-ref", "id":safeID(doc, "index")}, "Index"))

    if len(findAll(config.dfnElementsSelector, doc)):
        addIndexOfLocallyDefinedTerms(doc, container)

    if len(list(doc.externalRefsUsed.keys())):
        addIndexOfExternallyDefinedTerms(doc, container)


def addIndexOfLocallyDefinedTerms(doc, container):
    appendChild(container,
                E.h3({"class":"no-num no-ref", "id":safeID(doc, "index-defined-here")}, "Terms defined by this specification"))

    indexEntries = defaultdict(list)
    for el in findAll(config.dfnElementsSelector, doc):
        if el.get('id') is None or el.get('data-dfn-type') is None:
            continue
        linkTexts = config.linkTextsFromElement(el)
        headingLevel = headingLevelOfElement(el) or "Unnumbered section"
        type = el.get('data-dfn-type')
        if type == "argument":
            # Don't generate index entries for arguments.
            continue
        if el.get('data-dfn-for') is not None:
            disambiguator = "{0} for {1}".format(el.get('data-dfn-type'), ', '.join(config.splitForValues(el.get('data-dfn-for'))))
        elif type == "dfn":
            disambiguator = "definition of"
        else:
            disambiguator = "({0})".format(el.get('data-dfn-type'))

        id = el.get('id')
        for linkText in linkTexts:
            entry = {
                'url':"#" + id,
                'label':"§" + headingLevel,
                'disambiguator':disambiguator
            }
            indexEntries[linkText].append(entry)

    # Now print the indexes
    indexHTML = htmlFromIndexTerms(indexEntries)
    appendChild(container, indexHTML)


def addExplicitIndexes(doc):
    # Explicit indexes can be requested for specs with <index spec="example-spec-1"></index>

    for el in findAll("index", doc):

        status = el.get('status')
        if status and status not in config.specStatuses:
            die("<index> has unknown value '{0}' for status. Must be {1}.", status, config.englishFromList(config.specStatuses), el=el)
            continue

        if el.get('type'):
            types = set(x.strip() for x in el.get('type').split(','))
            for t in types:
                if t not in config.dfnTypes:
                    die("Unknown type value '{0}' on {1}".format(t, outerHTML(el)), el=el)
                    types.remove(t)
        else:
            types = None

        if el.get('data-link-spec'):
            # Yes, this is dumb. Accidental over-firing of a shortcut attribute. >_<
            specs = set(x.strip() for x in el.get('data-link-spec').split(','))
            for s in list(specs):
                if s not in doc.refs.specs:
                    die("Unknown spec name '{0}' on {1}".format(s, outerHTML(el)), el=el)
                    specs.remove(s)
        else:
            specs = None

        if el.get('for'):
            fors = set(x.strip() for x in el.get('for').split(','))
        else:
            fors = None

        if el.get('export'):
            exportVal = el.get('export').lower().strip()
            if exportVal in ["yes", "y", "true", "on"]:
                export = True
            elif exportVal in ["no", "n", "false", "off"]:
                export = False
            else:
                die("Unknown export value '{0}' (should be boolish) on {1}".format(exportVal, outerHTML(el)), el=el)
                export = None
        else:
            export = None

        def disambiguator(ref):
            disambInfo = []
            if types is None or len(types) > 1:
                disambInfo.append(ref.type)
            if specs is None or len(specs) > 1:
                disambInfo.append("in " + ref.spec)
            if ref.for_:
                try:
                    disambInfo.append("for {0}".format(', '.join(x.strip() for x in ref.for_)))
                except:
                    # todo: The TR version of Position triggers this
                    pass
            return ", ".join(disambInfo)


        # Initial filter of the ref database according to the <index> parameters
        possibleRefs = []
        for ref in doc.refs.queryAllRefs(dedupURLs=False, latestOnly=False, ignoreObsoletes=False):
            ref.text = re.sub(r"\s{2,}", " ", ref.text.strip())
            if export is not None and ref.export != export:
                continue
            if specs is not None and ref.spec not in specs:
                continue
            if types is not None and ref.type not in types:
                continue
            if fors is not None and not (set(ref.for_) & fors):
                continue
            possibleRefs.append(ref)

        # Group entries by linking text,
        # ensuring no duplicate disambiguators.
        refsFromText = defaultdict(list)
        for ref in possibleRefs:
            for i,existingRef in enumerate(refsFromText[ref.text]):
                if disambiguator(existingRef) != disambiguator(ref):
                    continue
                # Whoops, found an identical entry.
                if existingRef.status != ref.status:
                    if status:
                        if existingRef.status == status:
                            # Existing entry matches stated status, do nothing and don't add it.
                            break
                        elif ref.status == status:
                            # New entry matches status, update and don't re-add it.
                            refsFromText[text][i] = ref
                            break
                    else:
                        # Default to preferring current specs
                        if existingRef.status == "current":
                            break
                        elif ref.status == "current":
                            refsFromText[ref.text][i] = ref
                            break
                else:
                    # Legit dupes. Shouldn't happen in a good spec, but whatever.
                    pass
            else:
                refsFromText[ref.text].append(ref)

        # Group entries by text/type/for,
        # then filter each group for obsolete/oldversions.
        refsFromTtf = defaultdict(list)
        for text,entries in refsFromText.items():
            for ref in entries:
                ttf = (text, ref.type, "".join(ref.for_) if ref.for_ else None)
                refsFromTtf[ttf].append(ref)
        filteredRefs = defaultdict(list)
        for ttf, refs in list(refsFromTtf.items()):
            refs = doc.refs.filterObsoletes(refs)
            refs = refUtils.filterOldVersions(refs)
            if refs:
                filteredRefs[ttf[0]].extend({"url":ref.url, "disambiguator":disambiguator(ref)} for ref in refs)

        appendChild(el, htmlFromIndexTerms(filteredRefs))
        el.tag = "div"
        removeAttr(el, "export", "for", "spec", "status", "type")


def htmlFromIndexTerms(entries):
    # entries: dict (preferably OrderedDict, if you want stability) of linkText=>{url, label, disambiguator}
    # label is used for the actual link (normally heading level), disambiguator is phrase to use when there are collisions

    def entryKey(x):
        return (
            # first group by approximating a human-friendly "nearness" of terms
            re.sub(r'[^a-z0-9]', '', x[0].lower()),
            # then within that, by case-ignoring exact text
            x[0].lower(),
            # and finally uniquely by exact text
            x[0]
        )

    entries = OrderedDict(sorted(entries.items(), key=entryKey))

    topList = E.ul({"class":"index"})
    for text, items in entries.items():
        if len(items) == 1:
            item = items[0]
            li = appendChild(topList,
                             E.li(
                                 E.a({"href":item['url']}, text),
                                 E.span(", in ", item['label']) if item.get('label') else ""))
        else:
            li = appendChild(topList, E.li(text))
            ul = appendChild(li, E.ul())
            for item in sorted(items, key=lambda x:x['disambiguator']):
                appendChild(ul,
                            E.li(
                                E.a({"href":item['url']}, item['disambiguator']),
                                E.span(", in ", item['label']) if item.get('label') else ""))
    return topList


def addIndexOfExternallyDefinedTerms(doc, container):
    if not doc.externalRefsUsed:
        return

    def makeLink(*contents):
        return E.span({"style":"color:initial"}, *contents)

    ul = E.ul({"class": "index"})
    # Gather all the <a href> in the document, for use in the dfn-panels
    elsFromHref = DefaultOrderedDict(list)
    for a in findAll("a", doc):
        href = a.get("href")
        if href is None:
            continue
        if href.startswith("#"):
            continue
        elsFromHref[href].append(a)
    for spec, refGroups in sorted(doc.externalRefsUsed.items(), key=lambda x:x[0].upper()):
        # ref.spec is always lowercase; if the same string shows up in biblio data,
        # use its casing instead.
        biblioRef = doc.refs.getBiblioRef(spec, quiet=True)
        if biblioRef:
            printableSpec = biblioRef.linkText
        else:
            printableSpec = spec
        attrs = {"data-lt":spec, "data-link-type":"biblio", "data-biblio-type":"informative", "data-okay-to-fail": "true"}
        specLi = appendChild(ul,
                             E.li(
                                 E.a(attrs, "[", printableSpec, "]"), " defines the following terms:"))
        termsUl = appendChild(specLi, E.ul())
        for text,refs in sorted(refGroups.items(), key=lambda x:x[0]):
            if len(refs) == 1:
                ref = list(refs.values())[0]
                link = makeLink(ref.text)
            else:
                for key,ref in sorted(refs.items(), key=lambda x:x[0]):
                    if key:
                        link = makeLink(ref.text, " ", E.small({}, "(for {0})".format(key)))
                    else:
                        link = makeLink(ref.text)
            appendChild(termsUl, E.li(link))
            dfnpanels.addExternalDfnPanel(link, ref, elsFromHref, doc)
    appendChild(container,
                E.h3({"class":"no-num no-ref", "id":safeID(doc, "index-defined-elsewhere")}, "Terms defined by reference"),
                ul)


def addPropertyIndex(doc):
    # Extract all the data from the propdef and descdef tables

    if len(findAll("table.propdef, table.descdef", doc)) == 0:
        return
    html = getFillContainer('property-index', doc=doc, default=True)
    if html is None:
        return

    appendChild(html,
                E.h2({"class":"no-num no-ref", "id":safeID(doc, "property-index")}, "Property Index"))

    def extractKeyValFromRow(tr, table):
        # Extract the key, minus the trailing :
        result = re.match(r'(.*):', textContent(row[0]).strip())
        if result is None:
            die("Propdef row headers must be a word followed by a colon. Got:\n{0}", textContent(row[0]).strip(), el=table)
            return '',''
        key = result.group(1).strip().capitalize()
        # Extract the value from the second cell
        val = textContent(row[1]).strip()
        return key, val
    # Extract propdef info
    props = []
    for table in findAll('table.propdef', doc):
        prop = {}
        names = []
        for row in findAll('tr', table):
            key, val = extractKeyValFromRow(row, table)
            if key == "Name":
                names = [textContent(x) for x in findAll('dfn', row[1])]
            else:
                prop[key] = val
        for name in names:
            tempProp = prop.copy()
            tempProp['Name'] = name
            props.append(tempProp)
    props.sort(key=lambda x:x["Name"])
    # Extract descdef info
    atRules = defaultdict(list)
    for table in findAll('table.descdef', doc):
        desc = {}
        names = []
        atRule = ""
        for row in findAll('tr', table):
            key, val = extractKeyValFromRow(row, table)
            if key == "Name":
                names = [textContent(x) for x in findAll('dfn', row[1])]
            elif key == "For":
                atRule = val
            else:
                desc[key] = val
        for name in names:
            tempDesc = desc.copy()
            tempDesc['Name'] = name
            atRules[atRule].append(tempDesc)
    for desc in atRules.values():
        desc.sort(key=lambda x:x["Name"])

    def createRow(prop, linkType, for_=None):
        attrs = {"data-link-type":linkType}
        if for_:
            attrs["data-link-for"] = for_
        return E.tr(
            E.th({"scope":"row"},
                 E.a(attrs, prop['Name'])),
            *[E.td(prop.get(column,"")) for column in columns[1:]])
    if len(props):
        # Set up the initial table columns for properties
        columns = ["Name", "Value", "Initial", "Applies to", "Inherited", "Percentages"]
        # Add any additional keys used in the document.
        allKeys = set()
        for prop in props:
            allKeys |= set(prop.keys())
        columns.extend(sorted(allKeys - set(columns)))
        # Create the table

        def formatColumnName(name):
            if name == "Inherited":
                return "Inh."
            if name == "Percentages":
                return "%ages"
            if name == "Animatable":
                return "Ani\xadmat\xadable"
            if name == "Animation type":
                return "Anim\xadation type"
            if name == "Computed value":
                return "Com\xadputed value"
            return name
        appendChild(html,
                    E.div({"class":"big-element-wrapper"},
                          E.table({"class":"index"},
                                  E.thead(
                              E.tr(
                                  *[E.th({"scope":"col"}, formatColumnName(column)) for column in columns])),
                                  E.tbody(
                              *[createRow(prop, "property") for prop in props]))))
    else:
        appendChild(html, E.p("No properties defined."))

    if len(atRules):
        atRuleNames = sorted(atRules.keys())
        for atRuleName in atRuleNames:
            descs = atRules[atRuleName]
            if atRuleName == "":
                atRuleName = "Miscellaneous"
            columns = ["Name", "Value", "Initial"]
            allKeys = set()
            for desc in descs:
                allKeys |= set(desc.keys())
            columns.extend(sorted(allKeys - set(columns)))
            id = config.simplifyText(atRuleName) + "-descriptor-table"
            if atRuleName:
                appendChild(html,
                            E.h3({"class":"no-num no-ref", "id":safeID(doc, id)},
                                 E.a({"data-link-type":"at-rule"}, atRuleName),
                                 " Descriptors"))
            else:
                appendChild(html,
                            E.h3({"class":"no-num no-ref", "id":safeID(doc, id)},
                                 "Miscellaneous Descriptors"))
            appendChild(html,
                        E.div({"class":"big-element-wrapper"},
                              E.table({"class":"index"},
                                E.thead(
                                  E.tr(
                                      *[E.th({"scope":"col"}, column) for column in columns])),
                                E.tbody(
                                  *[createRow(desc, "descriptor", for_=atRuleName) for desc in descs]))))


def addIDLSection(doc):
    idlBlocks = [x for x in findAll("pre.idl, xmp.idl", doc) if isNormative(x, doc)]
    if len(idlBlocks) == 0:
        return
    html = getFillContainer('idl-index', doc=doc, default=True)
    if html is None:
        return

    appendChild(html,
                E.h2({"class":"no-num no-ref", "id":safeID(doc, "idl-index")}, "IDL Index"))

    container = appendChild(html, E.pre({"class":"idl"}))
    for block in idlBlocks:
        if hasClass(block, "extract"):
            continue
        blockCopy = copy.deepcopy(block)
        appendContents(container, blockCopy)
        appendChild(container, "\n")
    for el in findAll("[id]", container):
        if el.tag == "dfn":
            el.tag = "a"
            el.set("href", "#" + el.get("id"))
        del el.attrib["id"]
    addClass(container, "highlight")


def addTOCSection(doc):
    toc = getFillContainer("table-of-contents", doc=doc, default=False)
    if toc is None:
        return
    appendChild(toc,
                E.h2({"class": "no-num no-toc no-ref", "id":safeID(doc, "contents")}, "Table of Contents"))

    # containers[n] holds the current <ol> for inserting each heading's <li> into.
    # containers[1] is initialized with something arbitrary
    # (anything non-int and non-None will do, to avoid tripping error-handling),
    # and containers[2] is initialized with the top-level <ol>
    # (because <h2>s and below are all that should show up in your doc)
    # and whenever we hit a heading of level N,
    # we append an <li> for it to containers[N],
    # and create a new <ol> at containers[N+1] to hold its child sections
    # (nested inside the <li>).
    # We'll clean up empty subsections at the end.
    # There's a containers[0] just to make the levels map directly to indexes,
    # and a containers[7] just to avoid IndexErrors when dealing with <h6>s.
    # Aside from the above initialization,
    # containers[N] is initialized to an int,
    # arbitrarily set to the index just for readability.
    # If a heading specifies it shouldn't be in the ToC,
    # it sets containers[N+1] to None
    # to indicate that no children should go there.
    previousLevel = 1
    containers = [0, 1, 2, 3, 4, 5, 6, 7]
    containers[1] = toc
    containers[2] = appendChild(containers[1], E.ol({"class":"toc", "role":"directory"}))
    for header in findAll('h2, h3, h4, h5, h6', doc):
        level = int(header.tag[-1])
        container = containers[level]
        if isinstance(container, int):
            # Saw a low-level heading without first seeing a higher heading.
            die("Saw an <h{0}> without seeing an <h{1}> first. Please order your headings properly.\n{2}", level, level - 1, outerHTML(header), el=header)
            return
        if level > previousLevel + 1:
            # Jumping two levels is a no-no.
            die("Heading level jumps more than one level, from h{0} to h{1}:\n  {2}", previousLevel, level, textContent(header).replace("\n", " "), el=header)
            return

        addToTOC = True
        if hasClass(header, "no-toc"):
            # Hit a no-toc, suppress the entire section.
            addToTOC = False
        elif container is None:
            addToTOC = False
        elif (level-1) > doc.md.maxToCDepth:
            addToTOC = False

        if addToTOC:
            li = appendChild(container,
                             E.li(
                                 E.a({"href":"#" + header.get('id')},
                                     E.span({"class":"secno"},header.get('data-level', '')),
                                     " ",
                                     copy.deepcopy(find(".content", header)))))
            containers[level + 1] = appendChild(li, E.ol({"class":"toc"}))
        else:
            containers[level + 1] = None
        previousLevel = level

    container = containers[1]
    for el in findAll(".content a, .content dfn", container):
        el.tag = "span"
        if "href" in el.attrib:
            del el.attrib["href"]
    for el in findAll(".content [id]", container):
        del el.attrib["id"]
    for el in findAll("ol:empty", container):
        removeNode(el)


def addSpecMetadataSection(doc):
    def printEditor(editor):
        dd = E.dd({"class":"editor p-author h-card vcard"})
        if editor['w3cid']:
            dd.attrib['data-editor-id'] = editor['w3cid']
        if editor['link']:
            appendChild(dd, E.a({"class":"p-name fn u-url url", "href":editor['link']}, editor['name']))
        elif editor['email']:
            appendChild(dd, E.a({"class":"p-name fn u-email email", "href":"mailto:" + editor['email']}, editor['name']))
        else:
            appendChild(dd, E.span({"class":"p-name fn"}, editor['name']))
        if editor['org']:
            if editor['orglink']:
                el = E.a({"class":"p-org org", "href":editor['orglink']}, editor['org'])
            else:
                el = E.span({"class":"p-org org"}, editor['org'])
            appendChild(dd, " (", el, ")")
        if editor['email'] and editor['link']:
            appendChild(dd,
                        " ",
                        E.a({"class":"u-email email", "href":"mailto:" + editor['email']}, editor['email']))
        return dd

    def printTranslation(tr):
        lang = tr['lang-code']
        # canonicalize the lang-code structure
        lang = lang.lower().replace("_", "-")
        name = tr['name']
        nativeName = tr['native-name']
        url = tr['url']
        missingInfo = False
        if name is None:
            if lang in doc.languages:
                name = doc.languages[lang]['name']
            else:
                missingInfo = True
        if nativeName is None:
            if lang in doc.languages:
                nativeName = doc.languages[lang]['native-name']
            else:
                missingInfo = True
        if missingInfo:
            warn("Bikeshed doesn't have all the translation info for '{0}'. Please add to bikeshed/spec-data/readonly/languages.json and submit a PR!", lang)
        if nativeName:
            return E.span({"title": name or lang},
                          E.a({"href": url, "hreflang": lang, "rel": "alternate", "lang": lang},
                              nativeName))
        elif name:
            return E.a({"href": url, "hreflang": lang, "rel": "alternate", "title": lang},
                       name)
        else:
            return E.a({"href": url, "hreflang": lang, "rel": "alternate"},
                       lang)

    def printPreviousVersion(v):
        if v['type'] == "url":
            return E.a({"href":v['value'], "rel":"prev"}, v['value'])
        else:
            if v['type'] == "from-biblio":
                key = v['value']
            else: # "from-biblio-implicit"
                key = doc.md.vshortname
            dated = doc.refs.getLatestBiblioRef(key)
            if not dated:
                die(f"While trying to generate a Previous Version line, couldn't find a dated biblio reference for {key}.")
                return
            return E.a({"href":dated.url, "rel":"prev"}, dated.url)


    md = DefaultOrderedDict(list)
    mac = doc.macros
    if 'version' in mac:
        md["This version"].append(E.a({"href":mac['version'], "class":"u-url"}, mac['version']))
    if doc.md.TR:
        md["Latest published version"].append(E.a({"href": doc.md.TR}, doc.md.TR))
    if doc.md.ED and doc.md.status in config.snapshotStatuses:
        md["Editor's Draft"].append(E.a({"href": doc.md.ED}, doc.md.ED))
    if doc.md.previousVersions:
        md["Previous Versions"] = [printPreviousVersion(ver) for ver in doc.md.previousVersions]
    if doc.md.versionHistory:
        md["Version History"] = [E.a({"href":vh}, vh) for vh in doc.md.versionHistory]
    if doc.md.mailingList:
        span = E.span(
            E.a({"href":"mailto:" + doc.md.mailingList + "?subject=%5B" + mac['shortname'] + "%5D%20YOUR%20TOPIC%20HERE"}, doc.md.mailingList),
            " with subject line “",
            E.kbd(
                "[",
                mac['shortname'],
                "] ",
                E.i({"lt":""}, "… message topic …")),
            "”")
        if doc.md.mailingListArchives:
            appendChild(span,
                        " (",
                        E.a({"rel":"discussion", "href":doc.md.mailingListArchives}, "archives"),
                        ")")
        md["Feedback"].append(span)
    if doc.md.testSuite is not None:
        md["Test Suite"].append(E.a({"href":doc.md.testSuite}, doc.md.testSuite))
    elif (doc.md.vshortname in doc.testSuites) and (doc.testSuites[doc.md.vshortname]['url'] is not None):
        url = doc.testSuites[doc.md.vshortname]['url']
        md["Test Suite"].append(E.a({"href":url}, url))
    if doc.md.issues:
        md["Issue Tracking"] = [E.a({"href":href}, text) for text,href in doc.md.issues]
    if doc.md.editors:
        editorTerm = doc.md.editorTerm['singular']
        md[editorTerm] = list(map(printEditor, doc.md.editors))
    if doc.md.previousEditors:
        editorTerm = doc.md.editorTerm['singular']
        md["Former " + editorTerm] = list(map(printEditor, doc.md.previousEditors))
    if doc.md.translations:
        md["Translations"] = list(map(printTranslation, doc.md.translations))
    if doc.md.audience:
        md["Audience"] = [", ".join(doc.md.audience)]
    if doc.md.toggleDiffs:
        md["Toggle Diffs"] = [E.label({"for": safeID(doc, "hidedel"), "id": safeID(doc, "hidedel-label")}, "Hide deleted text")]
        prependChild(doc.body,
                     E.input({"type": "checkbox",
                              "id": safeID(doc, "hidedel"),
                              "style": "display:none"}))
        doc.extraStyles['style-hidedel'] = """
            #hidedel:checked ~ del, #hidedel:checked ~ * del { display:none; }
            #hidedel ~ #hidedel-label::before, #hidedel ~ * #hidedel-label::before { content: "☐ "; }
            #hidedel:checked ~ #hidedel-label::before, #hidedel:checked ~ * #hidedel-label::before { content: "☑ "; }
        """

    def createMdEntry(key, vals):
        vals = list(filter(lambda x:x is not None, vals))
        if not vals:
            return []
        # Pluralize appropriate words
        pluralization = {
            "Previous Version": "Previous Versions",
            "Test Suite": "Test Suites",
            doc.md.editorTerm['singular']: doc.md.editorTerm['plural'],
            "Former " + doc.md.editorTerm['singular']: "Former " + doc.md.editorTerm['plural']
        }
        if len(vals) > 1 and key in pluralization:
            displayKey = pluralization[key]
        else:
            displayKey = key
        if key in ("Editor", "Former Editor"):
            # A bunch of Microformats stuff is preloaded on the <dd>s,
            # so this prevents code from genning an extra wrapper <dd>.
            return [E.dt({"class": "editor"}, displayKey, ":")] + vals
        elif key == "Translations":
            ret = [E.dt(displayKey, " ", E.small("(non-normative)"), ":")]
            ret += [E.dd({}, val) for val in vals]
            return ret
        else:
            return [E.dt(displayKey, ":")] + [E.dd({}, val) for val in vals]

    # Merge "custom" metadata into non-custom, when they match up
    otherMd = OrderedDict()
    for k, vs in doc.md.otherMetadata.items():
        if k in md:
            md[k].extend(parseHTML(doc.fixText(v)) for v in vs)
        else:
            otherMd[k] = [parseHTML(doc.fixText(v)) for v in vs]

    dl = E.dl()
    for key in doc.md.metadataOrder:
        if key == "*":
            # Do all the non-explicit non-custom keys
            for k,vs in md.items():
                if k in doc.md.metadataOrder:
                    # Handled explicitly, don't put in the * spot
                    continue
                if k not in doc.md.metadataInclude:
                    # Explicitly excluded
                    continue
                appendChild(dl, *createMdEntry(k, vs))
        elif key == "!*":
            # Do all the non-explicit custom keys
            for k,vs in otherMd.items():
                if k in doc.md.metadataOrder:
                    continue
                if k not in doc.md.metadataInclude:
                    continue
                appendChild(dl, *createMdEntry(k, vs))
        elif key not in doc.md.metadataInclude:
            # Key explicitly excluded
            continue
        elif key in md:
            appendChild(dl, *createMdEntry(key, md[key]))
        elif key in otherMd:
            appendChild(dl, *createMdEntry(key, otherMd[key]))
    fillWith('spec-metadata', E.div(dl), doc=doc)


def addReferencesSection(doc):
    if not doc.normativeRefs and not doc.informativeRefs:
        return
    container = getFillContainer('references', doc=doc, default=True)
    if container is None:
        return

    def formatBiblioTerm(linkText):
        '''
        If the term is all uppercase, leave it like that.
        If it's all lowercase, uppercase it.
        If it's mixed case, leave it like that.
        '''
        if linkText.islower():
            return linkText.upper()
        return linkText

    appendChild(container,
                E.h2({"class":"no-num no-ref", "id":safeID(doc, "references")}, "References"))

    normRefs = sorted(doc.normativeRefs.values(), key=lambda r: r.linkText.lower())
    if len(normRefs):
        dl = appendChild(container,
                         E.h3({"class":"no-num no-ref", "id":safeID(doc, "normative")}, "Normative References"),
                         E.dl())
        for ref in normRefs:
            id = "biblio-" + config.simplifyText(ref.linkText)
            appendChild(dl, E.dt({"id":safeID(doc, id), "data-no-self-link":""}, "[" + formatBiblioTerm(ref.linkText) + "]"))
            appendChild(dl, E.dd(*ref.toHTML()))

    informRefs = [x for x in sorted(doc.informativeRefs.values(), key=lambda r: r.linkText.lower()) if x.linkText not in doc.normativeRefs]
    if len(informRefs):
        dl = appendChild(container,
                         E.h3({"class":"no-num no-ref", "id":safeID(doc, "informative")}, "Informative References"),
                         E.dl())
        for ref in informRefs:
            id = "biblio-" + config.simplifyText(ref.linkText)
            appendChild(dl, E.dt({"id":safeID(doc, id), "data-no-self-link":""}, "[" + formatBiblioTerm(ref.linkText) + "]"))
            appendChild(dl, E.dd(*ref.toHTML()))


def addIssuesSection(doc):
    issues = findAll('.issue', doc)
    if len(issues) == 0:
        return
    container = getFillContainer('issues-index', doc=doc, default=True)
    if container is None:
        return

    appendChild(container,
                E.h2({"class":"no-num no-ref", "id":safeID(doc, "issues-index")}, "Issues Index"))
    container = appendChild(container,
                            E.div({"style":"counter-reset:issue"}))
    for issue in issues:
        el = copy.deepcopy(issue)
        el.tail = None
        if el.tag not in ("pre","xmp"):
            el.tag = "div"
        appendChild(container, el)
        appendChild(el,
                    E.a({"href":"#" + issue.get('id')}, " ↵ "))
    for idel in findAll("[id]", container):
        del idel.attrib['id']
    for dfnel in findAll(config.dfnElementsSelector, container):
        dfnel.tag = "span"



# -*- coding: utf-8 -*-


import io

from subprocess import Popen, PIPE

from ..html import *
from ..messages import *

def processTags(doc):
	for el in findAll("[data-span-tag]", doc):
		tag = el.get("data-span-tag")
		if tag not in doc.md.inlineTagCommands:
			die("Unknown inline tag '{0}' found:\n  {1}", tag, outerHTML(el), el=el)
			continue
		command = doc.md.inlineTagCommands[tag]
		p = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
		out,err = p.communicate(innerHTML(el).encode('utf-8'))
		try:
			out = out.decode("utf-8")
		except UnicodeDecodeError as e:
			die("When trying to process {0}, got invalid unicode in stdout:\n{1}", outerHTML(el), e, el=el)
		try:
			err = err.decode("utf-8")
		except UnicodeDecodeError as e:
			die("When trying to process {0}, got invalid unicode in stderr:\n{1}", outerHTML(el), e, el=el)
		if p.returncode:
			die("When trying to process {0}, got return code {1} and the following stderr:\n{2}", outerHTML(el), p.returncode, err, el=el)
			continue
		replaceContents(el, parseHTML(out))



class NextLiteral:
	def __init__(self, regex):
		self.regex = regex

class NextBody:
	def __init__(self, regex):
		self.regex = regex

class Success:
	def __init__(self, nodes, skips=None):
		self.skips = skips or []
		self.nodes = nodes

class Failure:
	pass


# -*- coding: utf-8 -*-

import re
from pygments.lexer import *
from pygments.token import *


class CSSLexer(RegexLexer):
    name = "CSS"
    aliases = ['css']
    filenames = ['*.css']
    flags = re.DOTALL

    tokens = {
        "root": [
            (r"", Text, "rules"),
        ],
        "comment": [
            (r"\*/", Comment, "#pop"),
            (r".", Comment)
        ],
        "at-rule": [
            include("values"),
            (r";", Punctuation, "#pop"),
            (r"{\s*}", Punctuation, "#pop"),
            (r"{(?=[^;}]*{)", Punctuation, ("#pop", "rules")),
            (r"{(?=[^{}]*;)", Punctuation, ("#pop", "decls")),
            (r"", Text, "#pop")
        ],
        "rules": [
            (r"/\*", Comment, "comment"),
            (r"\s+", Text),
            (r"@[\w-]+", Name, "at-rule"),
            (r"}", Punctuation, "#pop"),
            (r"([^{]+)({)", bygroups(Name.Tag, Punctuation), "decls"),
        ],
        "decls": [
            (r";", Punctuation),
            (r"@[\w-]+", Name, "at-rule"),
            (r"([\w-]+)\s*(:)", bygroups(Keyword, Punctuation)),
            include("values"),
            (r"}", Punctuation, "#pop"),
            (r".+", Text)
        ],
        "values": [
            (r"/\*", Comment, "comment"),
            (r"[(),/]", Punctuation),
            (r"([+-]?(?:\d+(?:\.\d+)?|\d*\.\d+)[eE][+-]?(?:\d+(?:\.\d+)?|\d*\.\d+))([a-zA-Z-]+|%)", bygroups(Literal.Number, Literal)),
            (r"[+-]?(?:\d+(?:\.\d+)?|\d*\.\d+)[eE][+-]?(?:\d+(?:\.\d+)?|\d*\.\d+)", Literal.Number),
            (r"([+-]?(?:\d+(?:\.\d+)?|\d*\.\d+))([a-zA-Z-]+|%)", bygroups(Literal.Number, Literal)),
            (r"[+-]?(?:\d+(?:\.\d+)?|\d*\.\d+)", Literal.Number),
            (r"(url)(\()([^)]*)(\))", bygroups(Name.Function, Punctuation, Literal.String, Punctuation)),
            (r"([\w-]+)(\()", bygroups(Name.Function, Punctuation)),
            (r"\"[^\"]*\"", Literal.String),
            (r"'[^']*'", Literal.String),
            (r"#?[\w-]+", Text),
            (r"\s+", Text)
        ]
    }
